import re
from ..base import WriterBase
from ..core.class_ import Class
from ..core.function import Function
from ..core.object import Object, AccessSpecifier
from .regex import RegexPatternCpp
from ..utils.error import Error


class Writer(WriterBase):

    def __init__(self, out_directory):
        WriterBase.__init__(self, out_directory)
        self.objects_cache = {}
        self.methods_cache = {}
        self.methods_cache_with_templates = []

    def write_class(self, cls):
        self.objects_cache = {}
        self.methods_cache = {}

        for member in cls.members:
            declaration, initialisation, static_initialization = self.write_object(member)
            self.objects_cache[member.name] = [declaration, initialisation, static_initialization]

        for method in cls.functions:
            hpp, cpp = self.write_function(method)
            self.methods_cache[method] = ([hpp, cpp])
            if method.template_types:
                self.methods_cache_with_templates.append(hpp)

        header, includes, forward_declarations, forward_declarations_out = self.write_hpp(cls)
        source = self.write_cpp(cls, includes, forward_declarations, forward_declarations_out)
        return [
            (self.get_filename(cls, 'h'), self.prepare_file(header)),
            (self.get_filename(cls, 'cpp'), self.prepare_file(source)),
        ]

    def write_object(self, obj):
        declaration = ''
        initialization = ''
        static_initialization = ''

        assert isinstance(obj.type, str)
        if self.current_class.type == 'class':
            declaration = self.write_member_declaration(obj)
        elif self.current_class.type == 'enum':
            declaration = self.write_member_enum_declaration(obj)

        if self.current_class.type == 'enum':
            pass
        elif obj.is_static:
            static_initialization = self.write_member_static_init(self.current_class, obj)
        else:
            initialization += self.write_member_initialization(obj)
        return declaration, initialization, static_initialization

    def write_function(self, method):
        return [self.write_function_hpp(method), self.write_function_cpp(method)]

    def write_hpp(self, cls):
        namespace = 'mg'
        class_name = cls.name
        includes, forward_declarations, forward_declarations_out = self.get_includes_for_header(cls)
        functions = ''
        if cls.type == 'enum':
            functions += '{name}(const BaseEnum& rhs):BaseEnum(rhs){{}}'.format(name=class_name)
            functions += '\nconst {name}& operator =(const BaseEnum& rhs) {{ this->value = rhs.operator int(); return *this; }}'.format(name=class_name)

        access = AccessSpecifier.public
        for method in cls.functions:
            if method.name == 'constructor':
                continue
            hpp = self.methods_cache[method][0]
            if access != method.access:
                functions += AccessSpecifier.to_string(method.access) + ':\n'
                access = method.access
            functions += hpp
        for method in self.methods_cache_with_templates:
            includes.update(self.get_includes_for_method(cls, method, includes))
        members = ''

        access = AccessSpecifier.public
        for member in cls.members:
            if access != member.access:
                members += AccessSpecifier.to_string(member.access) + ':\n'
                access = member.access
            declaration, _, _ = self.objects_cache[member.name]
            members += declaration + '\n'

        virtual = 'virtual ' if cls.has_virtual_table() else ''
        destructor = '{virtual}~{name}();'.format(virtual=virtual, name=cls.name) if cls.type != 'enum' else ''

        superclass = '' if not cls.superclasses else ' : public %s' % cls.superclasses[0].name

        includes_s = self.build_includes(cls, includes)
        forward_declarations_s = self.build_forward_declarations(forward_declarations)
        forward_declarations_out_s = self.build_forward_declarations(forward_declarations_out)

        constructor_args = ''
        if cls.constructor is not None:
            constructor_args = self.create_function_hpp_args_string(cls.constructor)

        header = HEADER.format(namespace=namespace,
                               class_name=class_name,
                               destructor=destructor,
                               includes=includes_s,
                               forward_declarations=forward_declarations_s,
                               forward_declarations_out=forward_declarations_out_s,
                               functions=functions,
                               members=members,
                               superclass=superclass,
                               constructor_args=constructor_args)
        return header, includes, forward_declarations, forward_declarations_out

    def write_cpp(self, cls, includes, forw_declarations, forw_declarations_out):
        namespace = 'mg'
        class_name = cls.name
        functions = ''
        for method in cls.functions:
            if method.name == 'constructor':
                continue
            cpp = self.methods_cache[method][1]
            functions += cpp
        static_initializations = ''
        initializations = ''
        for member in cls.members:
            _, initialisation, static_initialization = self.objects_cache[member.name]
            if initialisation:
                div = ': ' if not initializations else ', '
                initializations += div + initialisation + '\n'
            if static_initialization:
                static_initializations += static_initialization + '\n'

        destructor = '''{name}::~{name}()
        {{
        }}
        '''.format(name=cls.name) if cls.type != 'enum' else ''

        registration = 'REGISTRATION_OBJECT({});\n'.format(cls.name) if self.model.auto_registration else ''
        is_abstract = cls.is_abstract
        if not is_abstract:
            for method in cls.functions:
                if method.is_abstract:
                    is_abstract = True
                    break
        if is_abstract:
            registration = ''
        if registration and not cls.has_method_with_name('get_type'):
            registration = ''

        includes = self.get_includes_for_source(cls, functions, includes, forw_declarations, forw_declarations_out)

        constructor_args = ''
        constructor_body = ''
        if cls.constructor is not None:
            constructor_args = self.create_function_cpp_args_string(cls.constructor)
            constructor_body = cls.constructor.body

        return SOURCE.format(namespace=namespace,
                             path_to_root=Writer.get_path_to_root(cls),
                             class_name=class_name,
                             destructor=destructor,
                             includes=includes,
                             functions=functions,
                             static_initializations=static_initializations,
                             initializations=initializations,
                             registration=registration,
                             constructor_args=constructor_args,
                             constructor_body=constructor_body)

    def create_function_hpp_args_string(self, method):
        args = list()
        for arg in method.args:
            assert isinstance(arg[1], Object)
            args.append(self.write_named_object(arg[1], arg[0], True, False))
            if arg[1].initial_value is not None:
                args[-1] += '=' + self.convert_initial_value(arg[1])
        args = ', '.join(args)
        return args

    def is_overrides(self, method: Function):
        if not method.is_virtual:
            return False
        superclass: Class = self.current_class.superclasses[0] if self.current_class.superclasses else None
        allowed = ['visit']
        while superclass is not None:
            for function in superclass.functions:
                if function.name != method.name:
                    continue
                if function.return_type.type != method.return_type.type:
                    if not self.model.validate_allow_different_virtual_method:
                        Error.exit(Error.ERROR_VIRTUAL_METHOD_HAS_DIFFERENT_DECLARATION, self.current_class.name,
                                   method.name, superclass.name)
                    continue

                if len(function.args) != len(method.args):
                    if not self.model.validate_allow_different_virtual_method and method.name not in allowed:
                        Error.exit(Error.ERROR_VIRTUAL_METHOD_HAS_DIFFERENT_DECLARATION, self.current_class.name,
                                   method.name, superclass.name)
                    continue
                cont = False
                for i, args in enumerate(function.args):
                    if args[1].type != method.args[i][1].type:
                        if not self.model.validate_allow_different_virtual_method and method.name not in allowed:
                            Error.exit(Error.ERROR_VIRTUAL_METHOD_HAS_DIFFERENT_DECLARATION, self.current_class.name,
                                       method.name, superclass.name)
                        cont = True
                        break
                if cont:
                    continue
                return True

            superclass = superclass.superclasses[0] if superclass.superclasses else None
        return False

    def write_function_hpp(self, method):
        string = '{virtual}{static}{friend}{type} {name}({args}){const}{override}{abstract}'
        assert isinstance(method.return_type, Object)
        return_type = self.write_named_object(method.return_type, '', False, True)
        args = self.create_function_hpp_args_string(method)
        virtual = 'virtual ' if method.is_virtual or method.is_abstract or self.current_class.is_virtual else ''

        body = ''
        if method.template_types:
            body = method.body
            string += '''
            {{
            {body}
            }}
            '''
            templates = ['class %s' % x for x in method.template_types]
            templates = ', '.join(templates)
            templates = 'template<%s> ' % templates
            string = templates + string
        else:
            string += ';\n'

        is_overrides = self.is_overrides(method)

        return string.format(virtual=virtual,
                             static='static ' if method.is_static else '',
                             friend='friend ' if method.is_friend else '',
                             const=' const' if method.is_const else '',
                             abstract=' = 0' if method.is_abstract else '',
                             override=' override' if is_overrides else '',
                             type=return_type,
                             name=method.name,
                             args=args,
                             body=body)

    def create_function_cpp_args_string(self, method):
        args = list()
        for arg in method.args:
            assert isinstance(arg[1], Object)
            args.append(self.write_named_object(arg[1], arg[0], True, False))
        args = ', '.join(args)
        return args

    def write_function_cpp(self, method):
        if method.is_external or method.is_abstract:
            return ''
        if method.template_types:
            return ''
        if method.specific_implementations:
            return method.specific_implementations

        scope = (self.current_class.name + '::') if not method.is_friend else ''
        text = '''{type} {scope}{name}({args}){const}
        {{
        {body}
        }}
        
        '''
        return_type = self.write_named_object(method.return_type, '', False, True)
        args = self.create_function_cpp_args_string(method)

        body = method.body

        return text.format(const=' const' if method.is_const else '',
                           type=return_type,
                           name=method.name,
                           args=args,
                           scope=scope,
                           body=body)

    def write_member_declaration(self, obj):
        return self.write_named_object(obj, obj.name, False, True) + ';'

    def write_member_enum_declaration(self, obj):
        if obj.name == 'value':
            return self.write_member_declaration(obj)
        return 'static constexpr BaseEnum {name} = {value};'.format(name=obj.name,
                                                                    value=obj.initial_value)

    def write_member_static_init(self, cls, obj, use_intrusive=True):
        string_pointer = '{const}intrusive_ptr<{type}{templates}> {owner}::{name}{initial_value};'
        string_non_pointer = '{const}{type}{templates}{pointer} {owner}::{name}{initial_value};'

        if use_intrusive and obj.is_pointer and not obj.is_const and not obj.is_link and not obj.denied_intrusive:
            string = string_pointer
        else:
            string = string_non_pointer

        templates = Writer.get_templates(obj)
        initial_value = self.convert_initial_value(obj)
        if initial_value:
            initial_value = f'({initial_value})'
        return string.format(const='const ' if obj.is_const else '',
                             type=self.convert_type(obj.type),
                             templates=templates,
                             pointer='*' if obj.is_pointer else '',
                             owner=cls.name,
                             name=obj.name,
                             initial_value=initial_value)

    def write_member_static_enum(self, cls, obj):
        assert self is not None
        return 'const int {}::{};'.format(cls.name, obj.name)

    def convert_initial_value(self, object_):
        if object_.is_pointer and (object_.initial_value == '0' or object_.initial_value is None):
            return 'nullptr'
        type_class = self.model.get_class(object_.type) if self.model.has_class(object_.type) else None
        if type_class is not None and type_class.type == 'enum' and object_.initial_value is None:
            assert type_class.members
            return '{}::{}'.format(type_class.name, type_class.members[0].name)

        if object_.initial_value is None:
            return ''
        return object_.initial_value

    def write_member_initialization(self, obj):
        initial_value = self.convert_initial_value(obj)
        return '{}({})'.format(obj.name, initial_value)

    @staticmethod
    def get_templates(obj):
        templates = []
        for arg in obj.template_args:
            assert isinstance(arg, Object)
            templates.append(Writer.write_named_object(arg, '', False, True))
            if arg.callable_args is not None:
                callable_args = []
                for callable_arg in arg.callable_args:
                    callable_args.append(Writer.write_named_object(callable_arg, '', True, False))
                callable_args = '({})'.format(', '.join(callable_args))
                templates[-1] += callable_args

        templates = ('<' + ', '.join(templates) + '>') if templates else ''
        return templates

    @staticmethod
    def write_named_object(obj, name, try_to_use_const_ref, use_intrusive):

        def can_use_const_ref(object_):
            assert isinstance(object_, Object)
            return object_.type in ['string', 'list', 'map']

        string_non_pointer = '{static}{const}{type}{templates}{pointer}{ref}{name}'
        string_pointer = '{static}{const}intrusive_ptr<{type}{templates}>{ref}{name}'
        templates = Writer.get_templates(obj)
        is_ref = obj.is_ref
        is_const = obj.is_const
        if not is_ref and not is_const and try_to_use_const_ref and can_use_const_ref(obj):
            is_ref = True
            is_const = True

        modified_type = obj.type
        if '(' in modified_type and ')' in modified_type:
            left = modified_type.index('(')
            right = modified_type.index(')')
            args = modified_type[left + 1:right]
            args = args.split(',')
            args = [Writer.convert_type(x.strip()) for x in args]
            args = ', '.join(args)
            modified_type = modified_type[0:left + 1] + args + modified_type[right:]

        if use_intrusive and obj.is_pointer and not obj.is_const and not obj.is_link and not obj.denied_intrusive:
            string = string_pointer
        else:
            string = string_non_pointer
        return string.format(static='static ' if obj.is_static else '',
                             const='const ' if is_const else '',
                             type=Writer.convert_type(modified_type),
                             name=(' ' + name) if name else '',
                             pointer='*' if obj.is_pointer else '',
                             ref='&' if is_ref else '',
                             templates=templates)

    @staticmethod
    def convert_type(type_of_object):
        assert isinstance(type_of_object, str)
        types = {
            'list': 'std::vector',
            'map': 'std::map',
            'string': 'std::string',
            'Observer': 'Observer<std::function<void()>>',
        }
        if type_of_object in types:
            return types[type_of_object]
        return type_of_object

    def prepare_file(self, text):
        tabs = 0

        lines = text.split('\n')
        text = list()

        def get_tabs(count):
            return '    ' * count

        for line in lines:
            line = line.strip()

            backward = False
            if line and line[0] == '}':
                tabs -= 1
            elif 'public:' in line or 'protected:' in line or 'private:' in line:
                backward = True
                tabs -= 1
            line_strip = line
            if line:
                line = get_tabs(tabs) + line
            if backward:
                tabs += 1
            if line_strip and line_strip[0] == '{':
                tabs += 1
            text.append(line)
        text = '\n'.join(text)
        text = text.strip().replace('\n\n\n', '\n\n') + '\n'
        return WriterBase.prepare_file(self, text)

    @staticmethod
    def get_filename(class_, ext):
        filename = class_.name + '.' + ext
        if class_.group:
            filename = class_.group + '/' + filename
        return filename

    def get_includes_for_header(self, cls):
        includes = set()
        forward_declarations = set()
        forward_declarations_out = set()

        def add(set_, obj):
            set_.add(self.convert_type(obj.type))
            for arg in obj.template_args:
                add(forward_declarations, arg)
            if obj.callable_args is not None:
                for arg in obj.callable_args:
                    add(forward_declarations, arg)

        # members
        for member in cls.members:
            def parse_object(container, obj):
                add(container, obj)
                for arg in obj.template_args:
                    if cls.name == 'DataStorage' or cls.prefer_use_forward_declarations:
                        parse_object(forward_declarations, arg)
                    else:
                        parse_object(container, arg)
            parse_object(includes, member)

        # functions
        std_includes = ['map', 'list', 'string']
        for method in cls.functions:
            for _, argtype in method.args:
                if argtype.type in std_includes:
                    add(includes, argtype)
                elif 'pugi::' in argtype.type or 'Json::' in argtype.type:
                    add(forward_declarations_out, argtype)
                else:
                    add(forward_declarations, argtype)
            if not self.model.user_includes or method.return_type.type in std_includes:
                add(includes, method.return_type)
            else:
                add(forward_declarations, method.return_type)

        # superclasses
        for superclass in cls.superclasses:
            includes.add(superclass.name)

        if cls.name in includes:
            includes.remove(cls.name)
        if cls.name in forward_declarations:
            forward_declarations.remove(cls.name)
        if cls.name in forward_declarations_out:
            forward_declarations_out.remove(cls.name)

        return includes, forward_declarations, forward_declarations_out

    def get_includes_for_source(self, cls: Class, functions_text, hpp_includes, forw_declarations, forw_declarations_out):
        includes = set()
        includes.add(cls.name)
        includes.update(forw_declarations)
        includes.update(forw_declarations_out)

        includes.update(self.get_includes_for_method(cls, functions_text, hpp_includes))
        includes.update(cls.user_includes)

        return self.build_includes(cls, includes)

    def get_includes_for_method(self, cls, functions_text, hpp_includes):
        includes = list()

        for type_ in self.model.classes:
            name = type_.name
            if name not in hpp_includes and name in functions_text:
                pattern = None
                if name not in RegexPatternCpp.regs_class_names:
                    pattern = re.compile(r'\b{}\b'.format(name))
                    RegexPatternCpp.regs_class_names[name] = pattern
                pattern = pattern or RegexPatternCpp.regs_class_names[name]
                need = cls.name is name
                need = need or pattern.search(functions_text) is not None
                if need:
                    includes.append(name)
        return includes

    def build_includes(self, cls, includes):
        types = {
            'std::list': '<vector>',
            'std::vector': '<vector>',
            'std::map': '<map>',
            'std::set': '<set>',
            'std::string': '<string>',
            'Json::Value': '"jsoncpp/json.h"',
            'pugi::xml_node': '"pugixml/pugixml.hpp"',
            'Observer': '"Observer.h"',
            'intrusive_ptr': '"intrusive_ptr.h"',
        }
        result = []
        for typename in includes:
            assert isinstance(typename, str)
            if typename in types:
                result.append('#include %s' % types[typename])
                continue

            if self.model.has_class(typename):
                other_class = self.model.get_class(typename)
                include = '#include '
                include += self.get_include_path_to_class(cls, other_class)
                result.append(include)

        result = sorted(result)
        return '\n'.join(result)

    @staticmethod
    def get_include_path_to_class(cls_, to_cls):
        include = '"'
        if cls_.group and to_cls.group != cls_.group:
            include += '../'
        if to_cls.group and to_cls.group != cls_.group:
            include += to_cls.group + '/'
        include += to_cls.name + '.h"'
        return include

    @staticmethod
    def get_path_to_root(cls_):
        if cls_.group:
            return '../'
        return ''

    def build_forward_declarations(self, declarations):
        ignore = [
            'std::list',
            'std::vector',
            'std::map',
            'std::set',
            'std::string',
            'int',
            'bool',
            'float',
            'void',
            'Observable',
        ]
        predefined = {
            'Json::Value': 'namespace Json\n{\nclass Value;\n}\n',
            'pugi::xml_node': 'namespace pugi\n{\nclass xml_node;\n}\n',
        }
        result = []
        for declaration in declarations:
            if declaration in ignore:
                continue
            if declaration in predefined:
                result.append(predefined[declaration])
            elif self.model.has_class(declaration):
                result.append('class %s;' % declaration)
        result = sorted(result)
        return '\n'.join(result)


HEADER = '''#ifndef __{namespace}_{class_name}_h__
#define __{namespace}_{class_name}_h__

#include <cstdint>
#include "intrusive_ptr.h"
#include "pugixml/pugixml.hpp"
{includes}

{forward_declarations_out}
namespace {namespace}
{{
    class SerializerXml;
    class DeserializerXml;
    class SerializerJson;
    class DeserializerJson;
{forward_declarations}

    class {class_name}{superclass}
    {{
    public:
        {class_name}({constructor_args});
        {destructor}
{functions}
{members}
    }};
}} //namespace {namespace}

#endif //#ifndef __{namespace}_{class_name}_h__
'''

SOURCE = '''
#include "intrusive_ptr.h"
#include "{path_to_root}{namespace}_Factory.h"
{includes}
#include "{path_to_root}{namespace}_extensions.h"
#include "{path_to_root}SerializerJson.h"
#include "{path_to_root}SerializerXml.h"

namespace {namespace}
{{
    {static_initializations}
    {registration}
    {class_name}::{class_name}({constructor_args})
    {initializations}{{
    {constructor_body}
    }}
    
    {destructor}
    
{functions}}} //namespace {namespace}

'''



JSON_TESTS = {
  "map<int|std::vector<float>>": {
    "object": [{
      "key": 123,
      "value": [
        1,
        2
      ]
    }]
  },
  "map<int|int>": {
    "object": [{
      "key": 123,
      "value": 123
    }]
  },
  "map<int|bool>": {
    "object": [{
      "key": 123,
      "value": True
    }]
  },
  "map<int|float>": {
    "object": [{
      "key": 123,
      "value": 123.5
    }]
  },
  "map<int|std::string>": {
    "object": [{
      "key": 123,
      "value": "434312some_random"
    }]
  },
  "map<int|TestEnum>": {
    "object": [{
      "key": 123,
      "value": "value2"
    }]
  },
  "map<int|const DataUnit*>": {
    "object": [{
      "key": 123,
      "value": "unit1"
    }]
  },
  "map<int|AllTypesChildren>": {
    "object": [{
      "key": 123,
      "value": {}
    }]
  },
  "map<int|intrusive_ptr<AllTypesChildren>>": {
    "object": [{
      "key": 123,
      "value": {
        "type": "AllTypesChildren"
      }
    }]
  },
  "map<int|std::vector<int>>": {
    "object": [{
      "key": 123,
      "value": [
        1,
        2,
        3,
        4
      ]
    }]
  },
  "map<int|std::vector<std::vector<bool>>>": {
    "object": [{
      "key": 123,
      "value": [
        [
          True,
          False
        ],
        [
          False,
          True
        ]
      ]
    }]
  },
  "map<int|std::map<int, int>>": {
    "object": [{
      "key": 123,
      "value": [{
          "key": 1,
          "value": 2
        },
        {
          "key": 2,
          "value": 3
        }
      ]
    }]
  },
  "map<bool|int>": {
    "object": [{
      "key": True,
      "value": 123
    }]
  },
  "map<bool|bool>": {
    "object": [{
      "key": True,
      "value": True
    }]
  },
  "map<bool|float>": {
    "object": [{
      "key": True,
      "value": 123.5
    }]
  },
  "map<bool|std::string>": {
    "object": [{
      "key": True,
      "value": "434312some_random"
    }]
  },
  "map<bool|TestEnum>": {
    "object": [{
      "key": True,
      "value": "value2"
    }]
  },
  "map<bool|const DataUnit*>": {
    "object": [{
      "key": True,
      "value": "unit1"
    }]
  },
  "map<bool|AllTypesChildren>": {
    "object": [{
      "key": True,
      "value": {}
    }]
  },
  "map<bool|intrusive_ptr<AllTypesChildren>>": {
    "object": [{
      "key": True,
      "value": {
        "type": "AllTypesChildren"
      }
    }]
  },
  "map<bool|std::vector<int>>": {
    "object": [{
      "key": True,
      "value": [
        1,
        2,
        3,
        4
      ]
    }]
  },
  "map<bool|std::vector<std::vector<bool>>>": {
    "object": [{
      "key": True,
      "value": [
        [
          True,
          False
        ],
        [
          False,
          True
        ]
      ]
    }]
  },
  "map<bool|std::map<int, int>>": {
    "object": [{
      "key": True,
      "value": [{
          "key": 1,
          "value": 2
        },
        {
          "key": 2,
          "value": 3
        }
      ]
    }]
  },
  "map<float|int>": {
    "object": [{
      "key": 123.5,
      "value": 123
    }]
  },
  "map<float|bool>": {
    "object": [{
      "key": 123.5,
      "value": True
    }]
  },
  "map<float|float>": {
    "object": [{
      "key": 123.5,
      "value": 123.5
    }]
  },
  "map<float|std::string>": {
    "object": [{
      "key": 123.5,
      "value": "434312some_random"
    }]
  },
  "map<float|TestEnum>": {
    "object": [{
      "key": 123.5,
      "value": "value2"
    }]
  },
  "map<float|const DataUnit*>": {
    "object": [{
      "key": 123.5,
      "value": "unit1"
    }]
  },
  "map<float|AllTypesChildren>": {
    "object": [{
      "key": 123.5,
      "value": {}
    }]
  },
  "map<float|intrusive_ptr<AllTypesChildren>>": {
    "object": [{
      "key": 123.5,
      "value": {
        "type": "AllTypesChildren"
      }
    }]
  },
  "map<float|std::vector<int>>": {
    "object": [{
      "key": 123.5,
      "value": [
        1,
        2,
        3,
        4
      ]
    }]
  },
  "map<float|std::vector<std::vector<bool>>>": {
    "object": [{
      "key": 123.5,
      "value": [
        [
          True,
          False
        ],
        [
          False,
          True
        ]
      ]
    }]
  },
  "map<float|std::map<int, int>>": {
    "object": [{
      "key": 123.5,
      "value": [{
          "key": 1,
          "value": 2
        },
        {
          "key": 2,
          "value": 3
        }
      ]
    }]
  },
  "map<std::string|int>": {
    "object": [{
      "key": "434312some_random",
      "value": 123
    }]
  },
  "map<std::string|bool>": {
    "object": [{
      "key": "434312some_random",
      "value": True
    }]
  },
  "map<std::string|float>": {
    "object": [{
      "key": "434312some_random",
      "value": 123.5
    }]
  },
  "map<std::string|std::string>": {
    "object": [{
      "key": "434312some_random",
      "value": "434312some_random"
    }]
  },
  "map<std::string|TestEnum>": {
    "object": [{
      "key": "434312some_random",
      "value": "value2"
    }]
  },
  "map<std::string|const DataUnit*>": {
    "object": [{
      "key": "434312some_random",
      "value": "unit1"
    }]
  },
  "map<std::string|AllTypesChildren>": {
    "object": [{
      "key": "434312some_random",
      "value": {}
    }]
  },
  "map<std::string|intrusive_ptr<AllTypesChildren>>": {
    "object": [{
      "key": "434312some_random",
      "value": {
        "type": "AllTypesChildren"
      }
    }]
  },
  "map<std::string|std::vector<int>>": {
    "object": [{
      "key": "434312some_random",
      "value": [
        1,
        2,
        3,
        4
      ]
    }]
  },
  "map<std::string|std::vector<std::vector<bool>>>": {
    "object": [{
      "key": "434312some_random",
      "value": [
        [
          True,
          False
        ],
        [
          False,
          True
        ]
      ]
    }]
  },
  "map<std::string|std::map<int, int>>": {
    "object": [{
      "key": "434312some_random",
      "value": [{
          "key": 1,
          "value": 2
        },
        {
          "key": 2,
          "value": 3
        }
      ]
    }]
  },
  "map<TestEnum|int>": {
    "object": [{
      "key": "value2",
      "value": 123
    }]
  },
  "map<TestEnum|bool>": {
    "object": [{
      "key": "value2",
      "value": True
    }]
  },
  "map<TestEnum|float>": {
    "object": [{
      "key": "value2",
      "value": 123.5
    }]
  },
  "map<TestEnum|std::string>": {
    "object": [{
      "key": "value2",
      "value": "434312some_random"
    }]
  },
  "map<TestEnum|TestEnum>": {
    "object": [{
      "key": "value2",
      "value": "value2"
    }]
  },
  "map<TestEnum|const DataUnit*>": {
    "object": [{
      "key": "value2",
      "value": "unit1"
    }]
  },
  "map<TestEnum|AllTypesChildren>": {
    "object": [{
      "key": "value2",
      "value": {}
    }]
  },
  "map<TestEnum|intrusive_ptr<AllTypesChildren>>": {
    "object": [{
      "key": "value2",
      "value": {
        "type": "AllTypesChildren"
      }
    }]
  },
  "map<TestEnum|std::vector<int>>": {
    "object": [{
      "key": "value2",
      "value": [
        1,
        2,
        3,
        4
      ]
    }]
  },
  "map<TestEnum|std::vector<std::vector<bool>>>": {
    "object": [{
      "key": "value2",
      "value": [
        [
          True,
          False
        ],
        [
          False,
          True
        ]
      ]
    }]
  },
  "map<TestEnum|std::map<int, int>>": {
    "object": [{
      "key": "value2",
      "value": [{
          "key": 1,
          "value": 2
        },
        {
          "key": 2,
          "value": 3
        }
      ]
    }]
  },
  "map<const DataUnit*|int>": {
    "object": [{
      "key": "unit1",
      "value": 123
    }]
  },
  "map<const DataUnit*|bool>": {
    "object": [{
      "key": "unit1",
      "value": True
    }]
  },
  "map<const DataUnit*|float>": {
    "object": [{
      "key": "unit1",
      "value": 123.5
    }]
  },
  "map<const DataUnit*|std::string>": {
    "object": [{
      "key": "unit1",
      "value": "434312some_random"
    }]
  },
  "map<const DataUnit*|TestEnum>": {
    "object": [{
      "key": "unit1",
      "value": "value2"
    }]
  },
  "map<const DataUnit*|const DataUnit*>": {
    "object": [{
      "key": "unit1",
      "value": "unit1"
    }]
  },
  "map<const DataUnit*|AllTypesChildren>": {
    "object": [{
      "key": "unit1",
      "value": {}
    }]
  },
  "map<const DataUnit*|intrusive_ptr<AllTypesChildren>>": {
    "object": [{
      "key": "unit1",
      "value": {
        "type": "AllTypesChildren"
      }
    }]
  },
  "map<const DataUnit*|std::vector<int>>": {
    "object": [{
      "key": "unit1",
      "value": [
        1,
        2,
        3,
        4
      ]
    }]
  },
  "map<const DataUnit*|std::vector<std::vector<bool>>>": {
    "object": [{
      "key": "unit1",
      "value": [
        [
          True,
          False
        ],
        [
          False,
          True
        ]
      ]
    }]
  },
  "map<const DataUnit*|std::map<int, int>>": {
    "object": [{
      "key": "unit1",
      "value": [{
          "key": 1,
          "value": 2
        },
        {
          "key": 2,
          "value": 3
        }
      ]
    }]
  },
  "map<AllTypesChildren|int>": {
    "object": [{
      "key": {},
      "value": 123
    }]
  },
  "map<AllTypesChildren|bool>": {
    "object": [{
      "key": {},
      "value": True
    }]
  },
  "map<AllTypesChildren|float>": {
    "object": [{
      "key": {},
      "value": 123.5
    }]
  },
  "map<AllTypesChildren|std::string>": {
    "object": [{
      "key": {},
      "value": "434312some_random"
    }]
  },
  "map<AllTypesChildren|TestEnum>": {
    "object": [{
      "key": {},
      "value": "value2"
    }]
  },
  "map<AllTypesChildren|const DataUnit*>": {
    "object": [{
      "key": {},
      "value": "unit1"
    }]
  },
  "map<AllTypesChildren|AllTypesChildren>": {
    "object": [{
      "key": {},
      "value": {}
    }]
  },
  "map<AllTypesChildren|intrusive_ptr<AllTypesChildren>>": {
    "object": [{
      "key": {},
      "value": {
        "type": "AllTypesChildren"
      }
    }]
  },
  "map<AllTypesChildren|std::vector<int>>": {
    "object": [{
      "key": {},
      "value": [
        1,
        2,
        3,
        4
      ]
    }]
  },
  "map<AllTypesChildren|std::vector<std::vector<bool>>>": {
    "object": [{
      "key": {},
      "value": [
        [
          True,
          False
        ],
        [
          False,
          True
        ]
      ]
    }]
  },
  "map<AllTypesChildren|std::map<int, int>>": {
    "object": [{
      "key": {},
      "value": [{
          "key": 1,
          "value": 2
        },
        {
          "key": 2,
          "value": 3
        }
      ]
    }]
  },
  "map<intrusive_ptr<AllTypesChildren>|int>": {
    "object": [{
      "key": {
        "type": "AllTypesChildren"
      },
      "value": 123
    }]
  },
  "map<intrusive_ptr<AllTypesChildren>|bool>": {
    "object": [{
      "key": {
        "type": "AllTypesChildren"
      },
      "value": True
    }]
  },
  "map<intrusive_ptr<AllTypesChildren>|float>": {
    "object": [{
      "key": {
        "type": "AllTypesChildren"
      },
      "value": 123.5
    }]
  },
  "map<intrusive_ptr<AllTypesChildren>|std::string>": {
    "object": [{
      "key": {
        "type": "AllTypesChildren"
      },
      "value": "434312some_random"
    }]
  },
  "map<intrusive_ptr<AllTypesChildren>|TestEnum>": {
    "object": [{
      "key": {
        "type": "AllTypesChildren"
      },
      "value": "value2"
    }]
  },
  "map<intrusive_ptr<AllTypesChildren>|const DataUnit*>": {
    "object": [{
      "key": {
        "type": "AllTypesChildren"
      },
      "value": "unit1"
    }]
  },
  "map<intrusive_ptr<AllTypesChildren>|AllTypesChildren>": {
    "object": [{
      "key": {
        "type": "AllTypesChildren"
      },
      "value": {}
    }]
  },
  "map<intrusive_ptr<AllTypesChildren>|intrusive_ptr<AllTypesChildren>>": {
    "object": [{
      "key": {
        "type": "AllTypesChildren"
      },
      "value": {
        "type": "AllTypesChildren"
      }
    }]
  },
  "map<intrusive_ptr<AllTypesChildren>|std::vector<int>>": {
    "object": [{
      "key": {
        "type": "AllTypesChildren"
      },
      "value": [
        1,
        2,
        3,
        4
      ]
    }]
  },
  "map<intrusive_ptr<AllTypesChildren>|std::vector<std::vector<bool>>>": {
    "object": [{
      "key": {
        "type": "AllTypesChildren"
      },
      "value": [
        [
          True,
          False
        ],
        [
          False,
          True
        ]
      ]
    }]
  },
  "map<intrusive_ptr<AllTypesChildren>|std::map<int, int>>": {
    "object": [{
      "key": {
        "type": "AllTypesChildren"
      },
      "value": [{
          "key": 1,
          "value": 2
        },
        {
          "key": 2,
          "value": 3
        }
      ]
    }]
  },
  "map<int|std::vector<const DataUnit*>>": {
    "object": [{
      "key": 123,
      "value": [
        "unit1",
        "unit1"
      ]
    }]
  },
  "map<int|std::vector<intrusive_ptr<AllTypesChildren>>": {
    "object": [{
      "key": 123,
      "value": [{
        "type": "AllTypesChildren"
      },{
        "type": "AllTypesChildren"
      }]
    }]
  },

}


XML_TESTS = {
  "map<int|std::vector<AllTypesChildren>>": "<root><object><pair key='123'><value><item /><item /></value></pair></object></root>",
  "map<int|std::vector<const DataUnit*>>": "<root><object><pair key='123'><value><item value='unit1' /><item value='unit1' /></value></pair></object></root>",
  "map<int|std::vector<intrusive_ptr<AllTypesChildren>>": "<root><object><pair key='123'><value><AllTypesChildren /><AllTypesChildren /></value></pair></object></root>",
  "map<int|int>": "<root><object><pair key='123' value='123' /></object></root>",
  "map<int|bool>": "<root><object><pair key='123' value='true' /></object></root>",
  "map<int|float>": "<root><object><pair key='123' value='123.5' /></object></root>",
  "map<int|std::string>": "<root><object><pair key='123' value='434312some_random' /></object></root>",
  "map<int|TestEnum>": "<root><object><pair key='123' value='value2' /></object></root>",
  "map<int|const DataUnit*>": "<root><object><pair key='123' value='unit1' /></object></root>",
  "map<int|AllTypesChildren>": "<root><object><pair key='123'><value /></pair></object></root>",
  "map<int|intrusive_ptr<AllTypesChildren>>": "<root><object><pair key='123'><value type='AllTypesChildren' /></pair></object></root>",
  "map<int|std::vector<int>>": "<root><object><pair key='123'><value><item value='1' /><item value='2' /><item value='3' /><item value='4' /></value></pair></object></root>",
  "map<int|std::vector<std::vector<bool>>>": "<root><object><pair key='123'><value><item><item value='true' /><item /></item><item><item /><item value='true'/></item></value></pair></object></root>",
  "map<int|std::map<int, int>>": "<root><object><pair key='123'><value><pair key='1' value='2' /><pair key='2' value='3' /></value></pair></object></root>",
  "map<bool|int>": "<root><object><pair key='true' value='123' /></object></root>",
  "map<bool|bool>": "<root><object><pair key='true' value='true' /></object></root>",
  "map<bool|float>": "<root><object><pair key='true' value='123.5' /></object></root>",
  "map<bool|std::string>": "<root><object><pair key='true' value='434312some_random' /></object></root>",
  "map<bool|TestEnum>": "<root><object><pair key='true' value='value2' /></object></root>",
  "map<bool|const DataUnit*>": "<root><object><pair key='true' value='unit1' /></object></root>",
  "map<bool|AllTypesChildren>": "<root><object><pair key='true'><value /></pair></object></root>",
  "map<bool|intrusive_ptr<AllTypesChildren>>": "<root><object><pair key='true'><value type='AllTypesChildren' /></pair></object></root>",
  "map<bool|std::vector<int>>": "<root><object><pair key='true'><value><item value='1' /><item value='2' /><item value='3' /><item value='4' /></value></pair></object></root>",
  "map<bool|std::vector<std::vector<bool>>>": "<root><object><pair key='true'><value><item><item value='true' /><item /></item><item><item /><item value='true'/></item></value></pair></object></root>",
  "map<bool|std::map<int, int>>": "<root><object><pair key='true'><value><pair key='1' value='2' /><pair key='2' value='3' /></value></pair></object></root>",
  "map<float|int>": "<root><object><pair key='123.5' value='123' /></object></root>",
  "map<float|bool>": "<root><object><pair key='123.5' value='true' /></object></root>",
  "map<float|float>": "<root><object><pair key='123.5' value='123.5' /></object></root>",
  "map<float|std::string>": "<root><object><pair key='123.5' value='434312some_random' /></object></root>",
  "map<float|TestEnum>": "<root><object><pair key='123.5' value='value2' /></object></root>",
  "map<float|const DataUnit*>": "<root><object><pair key='123.5' value='unit1' /></object></root>",
  "map<float|AllTypesChildren>": "<root><object><pair key='123.5'><value /></pair></object></root>",
  "map<float|intrusive_ptr<AllTypesChildren>>": "<root><object><pair key='123.5'><value type='AllTypesChildren' /></pair></object></root>",
  "map<float|std::vector<int>>": "<root><object><pair key='123.5'><value><item value='1' /><item value='2' /><item value='3' /><item value='4' /></value></pair></object></root>",
  "map<float|std::vector<std::vector<bool>>>": "<root><object><pair key='123.5'><value><item><item value='true' /><item /></item><item><item /><item value='true' /></item></value></pair></object></root>",
  "map<float|std::map<int, int>>": "<root><object><pair key='123.5'><value><pair key='1' value='2' /><pair key='2' value='3' /></value></pair></object></root>",
  "map<std::string|int>": "<root><object><pair key='434312some_random' value='123' /></object></root>",
  "map<std::string|bool>": "<root><object><pair key='434312some_random' value='true' /></object></root>",
  "map<std::string|float>": "<root><object><pair key='434312some_random' value='123.5' /></object></root>",
  "map<std::string|std::string>": "<root><object><pair key='434312some_random' value='434312some_random' /></object></root>",
  "map<std::string|TestEnum>": "<root><object><pair key='434312some_random' value='value2' /></object></root>",
  "map<std::string|const DataUnit*>": "<root><object><pair key='434312some_random' value='unit1' /></object></root>",
  "map<std::string|AllTypesChildren>": "<root><object><pair key='434312some_random'><value /></pair></object></root>",
  "map<std::string|intrusive_ptr<AllTypesChildren>>": "<root><object><pair key='434312some_random'><value type='AllTypesChildren' /></pair></object></root>",
  "map<std::string|std::vector<int>>": "<root><object><pair key='434312some_random'><value><item value='1' /><item value='2' /><item value='3' /><item value='4' /></value></pair></object></root>",
  "map<std::string|std::vector<std::vector<bool>>>": "<root><object><pair key='434312some_random'><value><item><item value='true' /><item /></item><item><item /><item value='true' /></item></value></pair></object></root>",
  "map<std::string|std::map<int, int>>": "<root><object><pair key='434312some_random'><value><pair key='1' value='2' /><pair key='2' value='3' /></value></pair></object></root>",
  "map<TestEnum|int>": "<root><object><pair key='value2' value='123' /></object></root>",
  "map<TestEnum|bool>": "<root><object><pair key='value2' value='true' /></object></root>",
  "map<TestEnum|float>": "<root><object><pair key='value2' value='123.5' /></object></root>",
  "map<TestEnum|std::string>": "<root><object><pair key='value2' value='434312some_random' /></object></root>",
  "map<TestEnum|TestEnum>": "<root><object><pair key='value2' value='value2' /></object></root>",
  "map<TestEnum|const DataUnit*>": "<root><object><pair key='value2' value='unit1' /></object></root>",
  "map<TestEnum|AllTypesChildren>": "<root><object><pair key='value2'><value /></pair></object></root>",
  "map<TestEnum|intrusive_ptr<AllTypesChildren>>": "<root><object><pair key='value2'><value type='AllTypesChildren' /></pair></object></root>",
  "map<TestEnum|std::vector<int>>": "<root><object><pair key='value2'><value><item value='1' /><item value='2' /><item value='3' /><item value='4' /></value></pair></object></root>",
  "map<TestEnum|std::vector<std::vector<bool>>>": "<root><object><pair key='value2'><value><item><item value='true' /><item /></item><item><item /><item value='true' /></item></value></pair></object></root>",
  "map<TestEnum|std::map<int, int>>": "<root><object><pair key='value2'><value><pair key='1' value='2' /><pair key='2' value='3' /></value></pair></object></root>",
  "map<const DataUnit*|int>": "<root><object><pair key='unit1' value='123' /></object></root>",
  "map<const DataUnit*|bool>": "<root><object><pair key='unit1' value='true' /></object></root>",
  "map<const DataUnit*|float>": "<root><object><pair key='unit1' value='123.5' /></object></root>",
  "map<const DataUnit*|std::string>": "<root><object><pair key='unit1' value='434312some_random' /></object></root>",
  "map<const DataUnit*|TestEnum>": "<root><object><pair key='unit1' value='value2' /></object></root>",
  "map<const DataUnit*|const DataUnit*>": "<root><object><pair key='unit1' value='unit1' /></object></root>",
  "map<const DataUnit*|AllTypesChildren>": "<root><object><pair key='unit1'><value /></pair></object></root>",
  "map<const DataUnit*|intrusive_ptr<AllTypesChildren>>": "<root><object><pair key='unit1'><value type='AllTypesChildren' /></pair></object></root>",
  "map<const DataUnit*|std::vector<int>>": "<root><object><pair key='unit1'><value><item value='1' /><item value='2' /><item value='3' /><item value='4' /></value></pair></object></root>",
  "map<const DataUnit*|std::vector<std::vector<bool>>>": "<root><object><pair key='unit1'><value><item><item value='true' /><item /></item><item><item /><item value='true' /></item></value></pair></object></root>",
  "map<const DataUnit*|std::map<int, int>>": "<root><object><pair key='unit1'><value><pair key='1' value='2' /><pair key='2' value='3' /></value></pair></object></root>",
  "map<AllTypesChildren|int>": "<root><object><pair value='123'><key /></pair></object></root>",
  "map<AllTypesChildren|bool>": "<root><object><pair value='true'><key /></pair></object></root>",
  "map<AllTypesChildren|float>": "<root><object><pair value='123.5'><key /></pair></object></root>",
  "map<AllTypesChildren|std::string>": "<root><object><pair value='434312some_random'><key /></pair></object></root>",
  "map<AllTypesChildren|TestEnum>": "<root><object><pair value='value2'><key /></pair></object></root>",
  "map<AllTypesChildren|const DataUnit*>": "<root><object><pair value='unit1'><key /></pair></object></root>",
  "map<AllTypesChildren|AllTypesChildren>": "<root><object><pair><key /><value /></pair></object></root>",
  "map<AllTypesChildren|intrusive_ptr<AllTypesChildren>>": "<root><object><pair><key /><value type='AllTypesChildren' /></pair></object></root>",
  "map<AllTypesChildren|std::vector<int>>": "<root><object><pair><key /><value><item value='1' /><item value='2' /><item value='3' /><item value='4' /></value></pair></object></root>",
  "map<AllTypesChildren|std::vector<std::vector<bool>>>": "<root><object><pair><key /><value><item><item value='true' /><item /></item><item><item /><item value='true' /></item></value></pair></object></root>",
  "map<AllTypesChildren|std::map<int, int>>": "<root><object><pair><key /><value><pair key='1' value='2' /><pair key='2' value='3' /></value></pair></object></root>",
  "map<intrusive_ptr<AllTypesChildren>|int>": "<root><object><pair value='123'><key type='AllTypesChildren' /></pair></object></root>",
  "map<intrusive_ptr<AllTypesChildren>|bool>": "<root><object><pair value='true'><key type='AllTypesChildren' /></pair></object></root>",
  "map<intrusive_ptr<AllTypesChildren>|float>": "<root><object><pair value='123.5'><key type='AllTypesChildren' /></pair></object></root>",
  "map<intrusive_ptr<AllTypesChildren>|std::string>": "<root><object><pair value='434312some_random'><key type='AllTypesChildren' /></pair></object></root>",
  "map<intrusive_ptr<AllTypesChildren>|TestEnum>": "<root><object><pair value='value2'><key type='AllTypesChildren' /></pair></object></root>",
  "map<intrusive_ptr<AllTypesChildren>|const DataUnit*>": "<root><object><pair value='unit1'><key type='AllTypesChildren' /></pair></object></root>",
  "map<intrusive_ptr<AllTypesChildren>|AllTypesChildren>": "<root><object><pair><key type='AllTypesChildren' /><value /></pair></object></root>",
  "map<intrusive_ptr<AllTypesChildren>|intrusive_ptr<AllTypesChildren>>": "<root><object><pair><key type='AllTypesChildren' /><value type='AllTypesChildren' /></pair></object></root>",
  "map<intrusive_ptr<AllTypesChildren>|std::vector<int>>": "<root><object><pair><key type='AllTypesChildren' /><value><item value='1' /><item value='2' /><item value='3' /><item value='4' /></value></pair></object></root>",
  "map<intrusive_ptr<AllTypesChildren>|std::vector<std::vector<bool>>>": "<root><object><pair><key type='AllTypesChildren' /><value><item><item value='true' /><item /></item><item><item /><item value='true' /></item></value></pair></object></root>",
  "map<intrusive_ptr<AllTypesChildren>|std::map<int, int>>": "<root><object><pair><key type='AllTypesChildren' /><value><pair key='1' value='2' /><pair key='2' value='3' /></value></pair></object></root>",
}



import re


def smart_split(string, divider, symbols=('<(', '>)')):
    assert len(divider) == 1
    if not string:
        return []

    left = symbols[0]
    right = symbols[1]

    parts = []
    depth = 0
    start = 0
    for curr, char in enumerate(string):
        if start >= curr:
            continue

        if char == divider and depth == 0:
            parts.append(string[start:curr])
            start = curr + 1
        elif char in left:
            depth += 1
        elif char in right:
            depth -= 1
    parts.append(string[start:])
    return parts


def parse_object(obj, string):
    if not string:
        return obj

    # remove template
    templates = ''
    left = string.find('<')
    if left != -1:
        right = left
        size = len(string)
        i = left + 1
        counter = 1
        while i < size:
            if string[i] == '<':
                counter += 1
            if string[i] == '>':
                counter -= 1
            if counter == 0:
                right = i
                break
            i += 1
        if counter == 0 and right != left:
            templates = string[left+1:right]
            string = string[:left] + string[right+1:]

    args = re.search(r'\(.*\)', string)
    args = args.group(0) if args else None
    if args:
        string = string.replace(args, '')
    type_s = re.search(r'\w+[&\*]*', string).group(0)
    string = string[len(type_s):]
    while True:
        match = re.search(r':\w+', string)
        if not match:
            break
        modifier = match.group(0)
        string = string[len(modifier):]
        type_s += modifier

    name_s = string.strip()

    obj.type = type_s + (args if args else '')
    obj.name = name_s

    obj.template_args = smart_split(templates, ',')
    obj.template_args = [x.strip() for x in obj.template_args]

    return obj


def tests():
    from ..core.object import Object

    def test_0():
        string = 'int a'
        obj = parse_object(Object(), string)
        # print(obj.name)
        # print(obj.type)
        # print(obj.template_args)
        return obj.name == 'a' and obj.type == 'int' and obj.template_args == []

    def test_1():
        string = 'list<list<int>> a'
        obj = parse_object(Object(), string)
        # print(obj.name)
        # print(obj.type)
        # print(obj.template_args)
        return obj.name == 'a' and obj.type == 'list' and obj.template_args == ['list<int>']

    def test_2():
        string = 'map<int, list<int>> foo'
        obj = parse_object(Object(), string)
        # print(obj.name)
        # print(obj.type)
        # print(obj.template_args)
        return obj.name == 'foo' and obj.type == 'map' and obj.template_args == ['int', 'list<int>']

    def test_3():
        string = 'map<int, DataReward*:link> foo'
        obj = parse_object(Object(), string)
        # print(obj.name)
        # print(obj.type)
        # print(obj.template_args)
        return obj.name == 'foo' and obj.type == 'map' and obj.template_args == ['int', 'DataReward*:link']

    def test_4():
        string = 'list<int>'
        obj = parse_object(Object(), string)
        # print(obj.name)
        # print(obj.type)
        # print(obj.template_args)
        return obj.type == 'list' and obj.template_args == ['int']

    def test_5():
        string = 'int'
        obj = parse_object(Object(), string)
        # print(obj.name)
        # print(obj.type)
        # print(obj.template_args)
        return obj.name == '' and obj.type == 'int' and obj.template_args == []

    print(test_0())
    print(test_1())
    print(test_2())
    print(test_3())
    print(test_4())
    print(test_5())


if __name__ == '__main__':
    tests()



import sys
import inspect


class Color(object):

    red = '\033[31m'
    green = '\033[32m'
    orange = '\033[33m'
    light_grey = '\033[37m'
    end = '\033[0m'


class Log(object):
    use_colors = True
    disable_logs = False

    def __init__(self):
        pass

    @staticmethod
    def debug(msg):
        if not Log.disable_logs:
            print(Color.light_grey + msg + Color.end if Log.use_colors else msg)

    @staticmethod
    def error(msg):
        print(Color.red + msg + Color.end if Log.use_colors else msg)

    @staticmethod
    def warning(msg):
        if not Log.disable_logs:
            print(Color.orange + msg + Color.end if Log.use_colors else msg)

    @staticmethod
    def message(msg):
        if not Log.disable_logs:
            print(Color.green + msg + Color.end if Log.use_colors else msg)


class Error(object):

    def __init__(self):
        pass

    UNKNOWN_SERIALISED_TYPE = inspect.currentframe().f_lineno
    STATIS_MEMBER_SHOULD_HAVE_INITIALISATION = inspect.currentframe().f_lineno
    MAP_TWO_ARGS = inspect.currentframe().f_lineno
    DUBLICATE_METHODS = inspect.currentframe().f_lineno
    DUBLICATE_CLASS = inspect.currentframe().f_lineno
    UNKNOWN_SUPERCLASS = inspect.currentframe().f_lineno
    UNKNOWN_CLASS = inspect.currentframe().f_lineno
    CANNOT_FIND_CLASS_FOR_METHOD = inspect.currentframe().f_lineno
    CANNOT_FIND_CLASS_FOR_OBJECT = inspect.currentframe().f_lineno
    OBJECT_IS_KEY_OF_MAP = inspect.currentframe().f_lineno
    CANNOT_PARSE_XML = inspect.currentframe().f_lineno
    ENUM_CANNOT_BE_COMBINATED = inspect.currentframe().f_lineno
    INTERNAL_ERROR = inspect.currentframe().f_lineno
    WARNING_SYNTAX = inspect.currentframe().f_lineno
    UNKNOWN_DATA_TYPE = inspect.currentframe().f_lineno
    CLASS_HAVE_MORE_THAN_ONE_CONSTRUCTOR = inspect.currentframe().f_lineno
    PARSE_ERROR = inspect.currentframe().f_lineno
    TESTS_FAILED = inspect.currentframe().f_lineno
    CIRCULAR_REFERENCE = inspect.currentframe().f_lineno
    ERROR_CONST_MODIFIER = inspect.currentframe().f_lineno
    WARNING_TEST_CLASS_NOT_IMPLEMENT_METHOD = inspect.currentframe().f_lineno
    ERROR_VIRTUAL_METHOD_HAS_DIFFERENT_DECLARATION = inspect.currentframe().f_lineno

    texts = {
        UNKNOWN_SERIALISED_TYPE: '[{}] unknown serialized serialized format. Base type - [{}]',
        STATIS_MEMBER_SHOULD_HAVE_INITIALISATION: 'Static method have to have a initial value. [{}::{}]',
        MAP_TWO_ARGS: 'Map have to have 2 arguments. [{}::{}]',
        DUBLICATE_METHODS: 'duplication function in one class [{}::{}]',
        DUBLICATE_CLASS: 'Duplication classes [{}]',
        UNKNOWN_SUPERCLASS: 'Cannot find superclass class: {}<{}>',
        UNKNOWN_CLASS: 'Unknown class: {}',
        CANNOT_FIND_CLASS_FOR_METHOD: 'Cannot find class [{}] for method [{}]',
        CANNOT_FIND_CLASS_FOR_OBJECT: 'Cannot find class [{}] for object [{}]',
        OBJECT_IS_KEY_OF_MAP: 'Validate php feature: key of array cannot be object [{}::map<{}, {}> {}]',
        CANNOT_PARSE_XML: 'Error on parsing xml [{}]',
        ENUM_CANNOT_BE_COMBINATED: 'Enum member [{}::{}] cannot be initialed by [{}] value',
        INTERNAL_ERROR: 'Internal error',
        WARNING_SYNTAX: 'Syntax warning: Found symbol ";" in member declaration: [{}]',
        UNKNOWN_DATA_TYPE: 'Unknown data type [{}]->[{}]. please check configuration. File: [{}]',
        CLASS_HAVE_MORE_THAN_ONE_CONSTRUCTOR: 'A class [{}] can have one constructor',
        PARSE_ERROR: 'Parsing error',
        TESTS_FAILED: 'Tests failed',
        CIRCULAR_REFERENCE: 'Circular reference: [{}]',
        ERROR_CONST_MODIFIER: 'Const modifier should be declare as [Type:const]: {}',
        WARNING_TEST_CLASS_NOT_IMPLEMENT_METHOD: 'Test class [{}] has not implemented method [{}]',
        ERROR_VIRTUAL_METHOD_HAS_DIFFERENT_DECLARATION: 'Class [{}] has different declaration of method [{}]. See class [{}]',
    }

    @staticmethod
    def exit(*args):
        Log.error((' Error {}: ' + Error.texts[args[0]]).format(args[0], *args[1:]))
        sys.exit(args[0])

    @staticmethod
    def warning(*args):
        Log.warning((' Warning {}: ' + Error.texts[args[0]]).format(args[0], *args[1:]))



from ..base.model import SerializeFormat, Model
from ..core.class_ import Class
from ..core.function import Function
from ..core.object import Objects, Object
from ..utils.error import Error

SERIALIZATION = 0
DESERIALIZATION = 1


class SerializerBase(object):

    def __init__(self):
        self.serialize_protocol = {}
        self.model: Model or None = None
        self.current_class: Class or None = None

    def generate_methods(self, model: Model):
        self.model = model

        formats = SerializeFormat.get_all()
        for serialize_format, string_format in formats:
            if model.serialize_formats & serialize_format:
                model.parser.load_default_serialize_protocol(self.get_protocol(string_format))
                self.serialize_protocol = model.serialize_protocol
                for cls in model.classes:
                    if cls.type == 'enum':
                        continue
                    if cls.name == 'BaseEnum':
                        continue
                    self.current_class = cls
                    self.create_serialization_function(cls, SERIALIZATION, string_format)
                    self.create_serialization_function(cls, DESERIALIZATION, string_format)

    def create_serialization_function(self, cls, serialize_type, serialize_format):
        method = Function()
        method.name = ('serialize_' if serialize_type == SERIALIZATION else 'deserialize_') + serialize_format
        method.translated = True
        method.return_type = Objects.VOID
        method.is_virtual = len(cls.superclasses) > 0 or len(cls.subclasses) > 0
        method.is_const = serialize_type == SERIALIZATION
        for func in cls.functions:
            if func.name == method.name:
                return method

        method.args.append(self.get_serialization_function_args(serialize_type, serialize_format))
        body = ''
        if cls.superclasses and cls.type != 'enum' and cls.superclasses[0].type != 'enum':
            parent_call = self.get_parent_serialize_format()
            body += parent_call.format(cls.superclasses[0].name, method.name, method.args[0][0])
        for obj in cls.members:
            if obj.is_runtime:
                continue
            if obj.is_static:
                continue
            if obj.is_const and not obj.is_link:
                continue
            body += self.build_serialize_operation(obj, serialize_type, serialize_format)
        method.body = body.rstrip()
        cls.functions.append(method)
        return method

    # pylint: disable=no-self-use
    # pylint: disable=unused-argument
    def get_protocol(self, serialize_format):
        assert False and "override me"

    def get_parent_serialize_format(self):
        assert False and "override me"

    def get_serialization_function_args(self, serialize_type, serialize_format):
        assert False and "override me"

    def build_serialize_operation(self, obj, serialization_type, serialize_format):
        assert False and "override me"

    def build_map_serialization(self, obj_name, obj_template_args, serialization_type, serialize_format):
        assert False and "override me"

    def finalize_serialize_operation(self, string):
        assert False and "override me"

    def convert_initialize_value(self, value):
        assert False and "override me"
    # pylint: enable=no-self-use
    # pylint: enable=unused-argument

    def build_serialize_operation_(self, obj_name, obj_type, obj_value, serialization_type, obj_template_args,
                                   obj_is_pointer, owner, is_link, serialize_format):
        index = 0
        if obj_value is None:
            index = 1

        type_ = obj_type
        cls = self.model.get_class(type_) if self.model.has_class(type_) else None
        arg_0 = obj_template_args[0].type if obj_template_args else 'unknown_arg'
        if cls is not None and cls.type == 'enum':
            type_ = 'enum'
        elif obj_type not in self.model.simple_types and type_ != "list" and type_ != "map":
            if is_link:
                type_ = 'link'
            elif obj_is_pointer:
                type_ = "pointer"
            else:
                type_ = "serialized"
        elif obj_type in self.model.simple_types:
            type_ = obj_type
        else:
            if obj_template_args:
                if type_ == "map":
                    if len(obj_template_args) != 2:
                        Error.exit(Error.MAP_TWO_ARGS, cls.name, obj_name)
                    return self.build_map_serialization(obj_name, obj_template_args,
                                                        serialization_type, serialize_format)

                arg = obj_template_args[0]
                assert isinstance(arg, Object)
                assert isinstance(arg.type, str)
                arg_type = arg.type
                type_cls = self.model.get_class(arg.type) if self.model.has_class(arg.type) else None
                if arg.is_link:
                    type_ = 'list<link>'
                elif arg_type in self.model.simple_types:
                    type_ = "list<{}>".format(arg_type)
                    obj_type = arg_type
                elif arg.is_pointer:
                    type_ = "list<pointer>"
                elif type_cls is not None and type_cls.type == 'enum':
                    type_ = 'list<string>'
                    arg_0 = 'string'
                else:
                    type_ = "list<serialized>"
                    obj_type = arg_type

        obj_value = self.convert_initialize_value(obj_value)
        string = self.serialize_protocol[serialization_type][type_][index]
        string = string.format(field=obj_name,
                               type=obj_type,
                               default_value=obj_value,
                               owner=owner,
                               arg_0=arg_0,
                               format=serialize_format)

        return self.finalize_serialize_operation(string)



from ..base import WriterBase
from .serializer import Serializer


class Writer(WriterBase):

    def __init__(self, out_directory):
        WriterBase.__init__(self, out_directory)

    def write_class(self, cls):
        self.set_initial_values(cls)

        imports_in_static = []
        imports_in_ctr = []
        initialize_list = ''
        static_list = ''
        slots = []
        for obj in cls.members:
            if not obj.is_static:
                v, i = self.write_object(obj)
                initialize_list += '        ' + v + '\n'
                imports_in_ctr.extend(i)
                slots.append('"%s"' % obj.name)
            else:
                v, i = self.write_object(obj)
                static_list += '    ' + v + '\n'
                imports_in_static.extend(i)
        slots = '__slots__ = [' + ', '.join(slots) + ']'

        imports_in_ctr = [' ' * 8 + x for x in imports_in_ctr]
        imports_in_static = [' ' * 4 + x for x in imports_in_static]
        initialize_list = '\n'.join(imports_in_ctr) + '\n' + initialize_list
        static_list = '\n'.join(imports_in_static) + '\n' + static_list

        if not initialize_list.strip():
            initialize_list = ' ' * 8 + 'pass'

        functions = ''
        for method in cls.functions:
            text = self.write_function(method)
            functions += text

        imports = 'from .common import *\n'
        if not cls.type == 'enum' and cls.name != 'BaseEnum':
            imports += 'from .SerializerXml import SerializerXml\n'
            imports += 'from .DeserializerXml import DeserializerXml\n'
            imports += 'from .SerializerJson import SerializerJson\n'
            imports += 'from .DeserializerJson import DeserializerJson\n'
            imports += 'from .IntrusivePtr import *\n'
        init_superclass = ''
        name = cls.name
        if cls.superclasses:
            superclass_name = cls.superclasses[0].name
            imports += 'from .{0} import {0}'.format(cls.superclasses[0].name)
            init_superclass = '        {0}.__init__(self)'.format(cls.superclasses[0].name)
        else:
            superclass_name = 'object'
        name += '(' + superclass_name + ')'

        for obj in cls.members:
            type_class = self.model.get_class(obj.type) if self.model.has_class(obj.type) else None
            if type_class is not None and type_class.type == 'enum':
                imports += '\nfrom .{0} import {0}'.format(type_class.name)

        for import_line in imports.split('\n'):
            functions = functions.replace(import_line, '')
            initialize_list = initialize_list.replace(import_line, '')

        if not init_superclass and not initialize_list:
            init_superclass = '        pass'

        constructor_args, constructor_body = self.get_constructor_data(cls)
        out = PATTERN_FILE.format(name=name,
                                  initialize_list=initialize_list,
                                  functions=functions,
                                  imports=imports,
                                  init_superclass=init_superclass,
                                  static_list=static_list,
                                  slots=slots,
                                  constructor_args=constructor_args,
                                  constructor_body=constructor_body)
        return [
            ('%s.py' % cls.name, self.prepare_file(out))
        ]

    def write_object(self, obj):
        imports = []
        if obj.name == 'from':
            obj.name = 'from_'
        value = obj.initial_value
        if value is None and not obj.is_pointer:
            type_ = obj.type
            if type_ == "string":
                value = '""'
            if type_ == "int":
                value = "0"
            if type_ == "float":
                value = "0"
            if type_ in ["uint", 'unsigned']:
                value = "0"
            if type_ == "bool":
                value = "False"
            if type_ == "list":
                value = "[]"
            if type_ == "map":
                value = "{}"
            else:
                if self.model.has_class(obj.type):
                    value = obj.type + '()'
                    tabs = ' ' * (4 if obj.is_static else 8)
                    imports.append(f'from .{obj.type} import {obj.type}')
        if value and value.endswith('f'):
            value = value[0:-1] + '0'

        if obj.is_static:
            out = '{0} = {1}'
        else:
            out = 'self.{0} = {1}'
        out = out.format(obj.name, Serializer().convert_initialize_value(value))
        return out, imports

    def prepare_file(self, text):
        lines = text.split('\n')
        result = []
        tabs = False
        is_static = False
        for line in lines:
            if not line.split():
                continue
            if not line.startswith('    ') and tabs:
                result.extend(['', ''])
            elif line.startswith('    @'):
                is_static = True
                result.extend([''])
                result.append(line)
                continue
            elif not is_static:
                if line.startswith('class ') or line.startswith('def '):
                    result.extend(['', ''])
                elif line.strip().startswith('def '):
                    result.extend([''])
            result.append(line)
            if line.startswith('    '):
                tabs = True
            is_static = False
        result.append('')
        text = '\n'.join(result)
        text = '# -*- coding: utf-8 -*-\n' + text
        return WriterBase.prepare_file(self, text)

    def get_method_arg_pattern(self, obj):
        return '{name}={value}' if obj.initial_value is not None else '{name}'

    def get_method_pattern(self, method):
        return PATTERN_METHOD

    def get_required_args_to_function(self, method):
        if not method or not method.is_static:
            return 'self'
        return None

    def get_nullptr_string(self):
        return 'None'

    def add_static_modifier_to_method(self, text):
        return '    @staticmethod' + text


PATTERN_FILE = '''
import xml.etree.ElementTree as ET
import json
from .mg_extensions import *
{imports}
class {name}:
{static_list}
    {slots}
    def __init__({constructor_args}):
{init_superclass}
{initialize_list}
{constructor_body}
    def __hash__(self):
        return id(self)
{functions}'''

PATTERN_METHOD = '''
    def {name}({args}):
{body}
'''



CPP_XML = '''
#int
#serialize:
#with default value:
if($(FIELD) != $(DEFAULT_VALUE))
{
    xml.append_attribute("$(FIELD)").set_value($(FIELD));
}
#without default value:
xml.append_attribute("$(FIELD)").set_value($(FIELD));

#deserialize:
#with default value:
$(FIELD) = xml.attribute("$(FIELD)").as_int($(DEFAULT_VALUE));
#without default value:
$(FIELD) = xml.attribute("$(FIELD)").as_int();


#bool
#serialize:
#with default value:
if($(FIELD) != $(DEFAULT_VALUE))
{
    xml.append_attribute("$(FIELD)").set_value($(FIELD));
}
#without default value:
xml.append_attribute("$(FIELD)").set_value($(FIELD));

#deserialize:
#with default value:
$(FIELD) = xml.attribute("$(FIELD)").as_bool($(DEFAULT_VALUE));
#without default value:
$(FIELD) = xml.attribute("$(FIELD)").as_bool();


#float
#serialize:
#with default value:
if($(FIELD) != $(DEFAULT_VALUE))
{
    xml.append_attribute("$(FIELD)").set_value($(FIELD));
}
#without default value:
xml.append_attribute("$(FIELD)").set_value($(FIELD));

#deserialize:
#with default value:
$(FIELD) = xml.attribute("$(FIELD)").as_float($(DEFAULT_VALUE));
#without default value:
$(FIELD) = xml.attribute("$(FIELD)").as_float();


#string
#serialize:
#with default value:
if($(FIELD) != $(DEFAULT_VALUE))
{
    xml.append_attribute("$(FIELD)").set_value($(FIELD).c_str());
}
#without default value:
xml.append_attribute("$(FIELD)").set_value($(FIELD).c_str());

#deserialize:
#with default value:
$(FIELD) = xml.attribute("$(FIELD)").as_string($(DEFAULT_VALUE));
#without default value:
$(FIELD) = xml.attribute("$(FIELD)").as_string();


#serialized
#serialize:
$(FIELD).serialize_$(FORMAT)(xml.append_child("$(FIELD)"));
#deserialize:
$(FIELD).deserialize_$(FORMAT)(xml.child("$(FIELD)"));


#pointer
#serialize
if($(FIELD))
{
    auto child = xml.append_child("$(FIELD)");
    child.append_attribute("type").set_value($(FIELD)->get_type().c_str());
    $(FIELD)->serialize_$(FORMAT)(child);
}
#deserialize:
auto xml_$(FIELD) = xml.child("$(FIELD)");
if(xml_$(FIELD))
{
    std::string type = xml_$(FIELD).attribute("type").as_string();
    $(FIELD) = Factory::shared().build<$(TYPE)>(type);
    if($(FIELD))
    {
        $(FIELD)->deserialize_$(FORMAT)(xml_$(FIELD));
    }
}


#list<int>
#serialize:
auto arr_$(FIELD) = xml.append_child("$(FIELD)");
for(auto t : $(FIELD))
{
    arr_$(FIELD).append_child("item").append_attribute("value").set_value(t);
}
#deserialize:
auto arr_$(FIELD) = xml.child("$(FIELD)");
for(const auto& child : arr_$(FIELD))
{
    $(FIELD).push_back(child.attribute("value").as_int());
}

#list<bool>
#serialize:
auto arr_$(FIELD) = xml.append_child("$(FIELD)");
for(auto t : $(FIELD))
{
    arr_$(FIELD).append_child("item").append_attribute("value").set_value(t);
}
#deserialize:
auto arr_$(FIELD) = xml.child("$(FIELD)");
for(const auto& child : arr_$(FIELD))
{
    $(FIELD).push_back(child.attribute("value").as_bool());
}
#list<float>
#serialize:
auto arr_$(FIELD) = xml.append_child("$(FIELD)");
for(auto t : $(FIELD))
{
    arr_$(FIELD).append_child("item").append_attribute("value").set_value(t);
}
#deserialize:
auto arr_$(FIELD) = xml.child("$(FIELD)");
for(const auto& child : arr_$(FIELD))
{
    $(FIELD).push_back(child.attribute("value").as_float());
}
#list<string>
#serialize:
auto arr_$(FIELD) = xml.append_child("$(FIELD)");
for(const auto& t : $(FIELD))
{
    arr_$(FIELD).append_child("item").append_attribute("value").set_value(t.c_str());
}
#deserialize:
auto arr_$(FIELD) = xml.child("$(FIELD)");
for(const auto& child : arr_$(FIELD))
{
    $(FIELD).push_back(child.attribute("value").as_string(""));
}

#list<serialized>
#serialize:
auto arr_$(FIELD) = xml.append_child("$(FIELD)");
for(auto& t : $(FIELD))
{
    t.serialize_$(FORMAT)(arr_$(FIELD).append_child("item"));
}
#deserialize:
auto arr_$(FIELD) = xml.child("$(FIELD)");
for(auto child : arr_$(FIELD))
{
    $(FIELD).emplace_back();
    $(FIELD).back().deserialize_$(FORMAT)(child);
}


#list<pointer>
#serialize:
auto arr_$(FIELD) = xml.append_child("$(FIELD)");
for(auto& t : $(FIELD))
{
    t->serialize_$(FORMAT)(arr_$(FIELD).append_child(t->get_type().c_str()));
}
#deserialize:
auto arr_$(FIELD) = xml.child("$(FIELD)");
for(auto child : arr_$(FIELD))
{
    auto type = child.name();
    auto obj = Factory::shared().build<$(ARG_0)>(type);
    if(obj)
    {
        $(FIELD).push_back(obj);
        $(FIELD).back()->deserialize_$(FORMAT)(child);
    }
}


#link
#serialize:
if($(FIELD))
{ 
    xml.append_attribute("$(FIELD)").set_value($(FIELD)->name.c_str());
}
#deserialize:
auto name_$(FIELD) = xml.attribute("$(FIELD)").as_string();
$(FIELD) = DataStorage::shared().get<$(TYPE)>(name_$(FIELD));


#list<link>
#serialize:
auto arr_$(FIELD) = xml.append_child("$(FIELD)");
for(auto& t : $(FIELD))
{
    arr_$(FIELD).append_child("data").append_attribute("value").set_value(t->name.c_str());
}
#deserialize:
auto arr_$(FIELD) = xml.child("$(FIELD)");
for(auto& child : arr_$(FIELD))
{
    auto name = child.attribute("value").as_string("");
    $(FIELD).push_back(DataStorage::shared().get<$(ARG_0)>(name));
}

#map
#serialize:
auto map_$(FIELD) = xml.append_child("$(FIELD)");
for(auto& pair : $(FIELD))
{
    auto xml = map_$(FIELD).append_child("pair");
    auto& key = pair.first;
    auto& value = pair.second;
    $(KEY_SERIALIZE)
    $(VALUE_SERIALIZE)
}
#deserialize:
auto map_$(FIELD) = xml.child("$(FIELD)");
for(auto child : map_$(FIELD))
{
    auto xml = child;
    $(KEY)
    $(VALUE_TYPE) value;
    $(KEY_SERIALIZE)
    $(VALUE_SERIALIZE)
    $(FIELD)[key] = value;
}

#enum
#serialize:
xml.append_attribute("$(FIELD)").set_value($(FIELD).str().c_str());
#deserialize:
$(FIELD) = std::string(xml.attribute("$(FIELD)").as_string(""));

#list<enum>
#serialize:
auto arr_$(FIELD) = xml.append_child("$(FIELD)");
for(auto t : $(FIELD))
{
    arr_$(FIELD).append_child("item").append_attribute("value").set_value(t.str().c_str());
}
#deserialize:
auto arr_$(FIELD) = xml.child("$(FIELD)");
for(const auto& child : arr_$(FIELD))
{
    $(FIELD).emplace_back(child.attribute("value").as_string());
}
'''


CPP_JSON = '''

#int, bool, float, string
#serialize:
#with default value:
if($(FIELD) != $(DEFAULT_VALUE))
{
    $(NAMESPACE)::set(json,"$(FIELD)",$(FIELD));
}
#without default value:
$(NAMESPACE)::set(json,"$(FIELD)",$(FIELD));

#deserialize:
#with default value:
if(json.isMember("$(FIELD)"))
{
    $(FIELD) = $(NAMESPACE)::get<$(TYPE)>(json["$(FIELD)"]);
}
else
{
    $(FIELD) = $(DEFAULT_VALUE);
}
#without default value:
$(FIELD) = $(NAMESPACE)::get<$(TYPE)>(json["$(FIELD)"]);


#serialized
#serialize:
$(FIELD).serialize_$(FORMAT)(json["$(FIELD)"]);
#deserialize:
$(FIELD).deserialize_$(FORMAT)(json["$(FIELD)"]);


#pointer
#serialize
if($(FIELD))
{
    $(FIELD)->serialize_$(FORMAT)(json["$(FIELD)"][$(FIELD)->get_type()]);
}
#deserialize:
if(json.isMember("$(FIELD)"))
{
    auto type_$(FIELD) = json["$(FIELD)"].getMemberNames()[0];
    $(FIELD) = Factory::shared().build<$(TYPE)>(type_$(FIELD));
    if($(FIELD))
    {
        $(FIELD)->deserialize_$(FORMAT)(json["$(FIELD)"][type_$(FIELD)]);
    }
}


#list<bool>
#serialize:
auto& arr_$(FIELD) = json["$(FIELD)"];
int i_$(FIELD)=0;
for(const auto& t : $(FIELD))
{
    arr_$(FIELD)[i_$(FIELD)++] = Json::Value(t);
}
#deserialize:
auto& arr_$(FIELD) = json["$(FIELD)"];
for(int i = 0; i < arr_$(FIELD).size(); ++i)
{
    $(FIELD).push_back($(NAMESPACE)::get<$(ARG_0)>(arr_$(FIELD)[i]));
}

#list<int>, list<float>, list<string>
#serialize:
auto& arr_$(FIELD) = json["$(FIELD)"];
int i_$(FIELD)=0;
for(const auto& t : $(FIELD))
{
    $(NAMESPACE)::set(arr_$(FIELD)[i_$(FIELD)++], t);
}
#deserialize:
auto& arr_$(FIELD) = json["$(FIELD)"];
for(int i = 0; i < arr_$(FIELD).size(); ++i)
{
    $(FIELD).emplace_back();
    $(FIELD).back() = $(NAMESPACE)::get<$(ARG_0)>(arr_$(FIELD)[i]);
}


#list<serialized>
#serialize:
auto& arr_$(FIELD) = json["$(FIELD)"];
int i_$(FIELD)=0;
for(auto& t : $(FIELD))
{
    t.serialize_$(FORMAT)(arr_$(FIELD)[i_$(FIELD)++]);
}
#deserialize:
auto& arr_$(FIELD) = json["$(FIELD)"];
for(int i = 0; i < arr_$(FIELD).size(); ++i)
{
    $(FIELD).emplace_back();
    $(FIELD).back().deserialize_$(FORMAT)(arr_$(FIELD)[i]);
}


#list<pointer>
#serialize:
auto& arr_$(FIELD) = json["$(FIELD)"];
for(auto& t : $(FIELD))
{
    auto index = arr_$(FIELD).size();
    t->serialize_$(FORMAT)(arr_$(FIELD)[index][t->get_type()]);
}
#deserialize:
auto& arr_$(FIELD) = json["$(FIELD)"];
auto size_$(FIELD) = arr_$(FIELD).size();
for(int i = 0; i < size_$(FIELD); ++i)
{
    auto type = arr_$(FIELD)[i].getMemberNames()[0];
    auto obj = Factory::shared().build<$(ARG_0)>(type);
    if(obj)
    {
        $(FIELD).emplace_back(obj);
        $(FIELD).back()->deserialize_$(FORMAT)(arr_$(FIELD)[i][type]);
    }
}


#link
#serialize:
if($(FIELD))
{
    $(NAMESPACE)::set(json,"$(FIELD)",$(FIELD)->name);
}
#deserialize:
$(FIELD) = DataStorage::shared().get<$(TYPE)>($(NAMESPACE)::get<std::string>(json["$(FIELD)"]));


#list<link>
#serialize:
auto& arr_$(FIELD) = json["$(FIELD)"];
for(auto& item : $(FIELD))
{
    arr_$(FIELD).append(item->name);
}
#deserialize:
auto& arr_$(FIELD) = json["$(FIELD)"];
for(auto item : arr_$(FIELD))
{
    auto name = $(NAMESPACE)::get<std::string>(item);
    auto data = DataStorage::shared().get<$(ARG_0)>(name);
    $(FIELD).push_back(data);
}


#map
#serialize:
auto& map_$(FIELD) = json["$(FIELD)"];
for(auto& pair : $(FIELD))
{
    auto& json = map_$(FIELD)[map_$(FIELD).size()];
    auto& key = pair.first;
    auto& value = pair.second;
    $(KEY_SERIALIZE)
    $(VALUE_SERIALIZE)
}
#deserialize:
auto& map_$(FIELD) = json["$(FIELD)"];
auto size_$(FIELD)= map_$(FIELD).size();
for(unsigned int i = 0; i < size_$(FIELD); ++i)
{
    auto& json = map_$(FIELD)[i];
    $(KEY);
    $(VALUE_TYPE) value;
    $(VALUE_SERIALIZE)
    $(KEY_SERIALIZE)
    $(FIELD)[key] = value;
}


#enum
#serialize:
$(NAMESPACE)::set(json, "$(FIELD)", $(FIELD).str());
#deserialize:
$(FIELD) = $(NAMESPACE)::get<std::string>(json["$(FIELD)"]);

#list<enum>
#serialize:
auto& arr_$(FIELD) = json["$(FIELD)"];
int i_$(FIELD)=0;
for(const auto& t : $(FIELD))
{
    $(NAMESPACE)::set(arr_$(FIELD)[i_$(FIELD)++], t.str());
}
#deserialize:
auto& arr_$(FIELD) = json["$(FIELD)"];
for(int i = 0; i < arr_$(FIELD).size(); ++i)
{
    $(FIELD).emplace_back($(NAMESPACE)::get<std::string>(arr_$(FIELD)[i]));
}
'''



# -*- coding: utf-8 -*-
import xml.etree.ElementTree as ET
import json
from .common import *


class BaseEnum(object):
    TYPE = "BaseEnum"
    __slots__ = []

    def __init__(self):
        pass

    def __hash__(self):
        return id(self)

    def __eq__(self, rhs):
        result = True
        return result
        pass

    def __ne__(self, rhs):
        return not (self == rhs)
        pass

    def get_type(self):
        return BaseEnum.TYPE
        pass



from tests.test_serializer.py.gen.AllTypes import AllTypes
from tests.test_serializer.py.gen.IntrusivePtr import make_intrusive


def test_make_intrusive_with_args():
    assert make_intrusive(AllTypes).int_value0 == 0
    assert make_intrusive(AllTypes, 1).int_value0 == 1
    print('Ok: test_make_intrusive_with_args')


def test_compare_intrusive():
    a = make_intrusive(AllTypes)
    b = make_intrusive(AllTypes)
    assert a == b
    print('Ok: test_compare_intrusive')


def test_intrusive():
    test_make_intrusive_with_args()
    test_compare_intrusive()



import unittest

import os
import sys
import inspect

sys.path.insert(0, os.path.abspath(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))) + '/..'))
from mlc_tools.core.function import *
from mlc_tools.base.parser import Parser
from mlc_tools.base.model import Model


class TestParseObject(unittest.TestCase):

    def test_1(self):
        parser = Parser(Model())
        text = 'map<int, list<Foo>> some_map'
        parser.parse_text(text)
        obj = parser.model.objects[0]
        self.assertTrue(isinstance(obj, Object))
        self.assertEqual(obj.type, 'map')
        self.assertEqual(obj.name, 'some_map')
        self.assertTrue(len(obj.template_args), 2)
        self.assertTrue(isinstance(obj.template_args[0], Object))
        self.assertTrue(isinstance(obj.template_args[1], Object))
        self.assertEqual(obj.template_args[0].type, 'int')
        self.assertEqual(obj.template_args[0].name, '')
        self.assertEqual(obj.template_args[1].type, 'list')
        self.assertEqual(obj.template_args[1].name, '')
        self.assertTrue(len(obj.template_args[1].template_args), 1)
        self.assertTrue(isinstance(obj.template_args[1].template_args[0], Object))
        self.assertEqual(obj.template_args[1].template_args[0].type, 'Foo')
        self.assertEqual(obj.template_args[1].template_args[0].name, '')

    def test_observable_parse(self):
        parser = Parser(Model())

        text = 'void()'
        parser.parse_text(text)
        obj = parser.model.objects[0]
        self.assertEqual(obj.type, 'void')
        self.assertEqual(len(obj.callable_args), 0)

        parser.model.objects = []
        text = 'void(int)'
        parser.parse_text(text)
        obj = parser.model.objects[0]
        self.assertEqual(obj.type, 'void')
        self.assertEqual(len(obj.callable_args), 1)
        self.assertTrue(isinstance(obj.callable_args[0], Object))
        self.assertEqual(obj.callable_args[0].type, 'int')
        self.assertIsNotNone(obj.callable_args)

        parser.model.objects = []
        text = 'Observable<void()> eventTest'
        parser.parse_text(text)
        obj = parser.model.objects[0]
        self.assertEqual(obj.type, 'Observable')
        self.assertEqual(obj.name, 'eventTest')
        self.assertEqual(len(obj.template_args), 1)
        self.assertEqual(obj.template_args[0].type, 'void')
        self.assertEqual(len(obj.template_args[0].callable_args), 0)
        self.assertIsNotNone(obj.template_args[0].callable_args)

        parser.model.objects = []
        text = 'Observable<void(int, int)> eventTest'
        parser.parse_text(text)
        obj = parser.model.objects[0]
        self.assertEqual(obj.type, 'Observable')
        self.assertEqual(obj.name, 'eventTest')
        self.assertEqual(len(obj.template_args), 1)
        self.assertEqual(obj.template_args[0].type, 'void')
        self.assertEqual(len(obj.template_args[0].callable_args), 2)
        self.assertEqual(obj.template_args[0].callable_args[0].type, 'int')
        self.assertEqual(obj.template_args[0].callable_args[1].type, 'int')
        self.assertIsNotNone(obj.template_args[0].callable_args)

        parser.model.objects = []
        text = 'Observable<void(DataUnit:link)> eventTest'
        parser.parse_text(text)
        obj = parser.model.objects[0]
        self.assertEqual(obj.type, 'Observable')
        self.assertEqual(obj.name, 'eventTest')
        self.assertEqual(len(obj.template_args), 1)
        self.assertEqual(obj.template_args[0].type, 'void')
        self.assertEqual(len(obj.template_args[0].callable_args), 1)
        self.assertEqual(obj.template_args[0].callable_args[0].type, 'DataUnit')
        self.assertEqual(obj.template_args[0].callable_args[0].is_link, True)
        self.assertIsNotNone(obj.template_args[0].callable_args)

    def test_observable_parse_with_arg(self):
        parser = Parser(Model())
        text = 'void(int)'
        parser.parse_text(text)
        obj = parser.model.objects[0]
        self.assertEqual(obj.type, 'void')


class TestParseFunction(unittest.TestCase):

    def test_1(self):
        parser = Parser(Model())
        text = 'function map<int, list<Foo>> some_function(map<float, list<Bar>> arg0, list<int> arg1){}'
        parser.parse_text(text)
        func = parser.model.functions[0]
        self.assertTrue(isinstance(func, Function))
        self.assertEqual(func.name, 'some_function')
        self.assertEqual(len(func.args), 2)
        self.assertEqual(len(func.operations), 0)

        self.assertTrue(isinstance(func.return_type, Object))
        self.assertEqual(func.return_type.type, 'map')
        self.assertEqual(func.return_type.name, '')
        self.assertEqual(len(func.return_type.template_args), 2)
        self.assertTrue(isinstance(func.return_type.template_args[0], Object))
        self.assertTrue(isinstance(func.return_type.template_args[1], Object))
        self.assertEqual(func.return_type.template_args[0].name, '')
        self.assertEqual(func.return_type.template_args[0].type, 'int')
        self.assertEqual(func.return_type.template_args[1].name, '')
        self.assertEqual(func.return_type.template_args[1].type, 'list')
        self.assertEqual(len(func.return_type.template_args[1].template_args), 1)
        self.assertTrue(isinstance(func.return_type.template_args[1].template_args[0], Object))
        self.assertEqual(func.return_type.template_args[1].template_args[0].name, '')
        self.assertEqual(func.return_type.template_args[1].template_args[0].type, 'Foo')

        self.assertEqual(func.args[0][0], 'arg0')
        self.assertEqual(func.args[1][0], 'arg1')
        self.assertTrue(isinstance(func.args[0][1], Object))
        self.assertTrue(isinstance(func.args[1][1], Object))
        self.assertEqual(func.args[0][1].name, 'arg0')
        self.assertEqual(func.args[0][1].type, 'map')
        self.assertTrue(isinstance(func.args[0][1].template_args[0], Object))
        self.assertTrue(isinstance(func.args[0][1].template_args[1], Object))
        self.assertEqual(func.args[1][1].name, 'arg1')
        self.assertEqual(func.args[1][1].type, 'list')
        self.assertTrue(isinstance(func.args[0][1].template_args[0], Object))

    def test_2(self):
        text = 'function void visit(Request* ctx)'
        parser = Parser(Model())
        parser.parse_text(text)
        func = parser.model.functions[0]
        self.assertTrue(isinstance(func.return_type, Object))
        self.assertEqual(func.return_type.type, 'void')
        self.assertEqual(func.return_type.name, '')

    def test_3(self):
        parser = Parser(Model())
        parser.parse_text('function void initialize(ModelUser*user){}')
        func = parser.model.functions[0]
        self.assertEqual(func.args[0][0], 'user')
        self.assertEqual(func.args[0][1].type, 'ModelUser')

    def test_4(self):
        parser = Parser(Model())
        parser.parse_text('function DataAbility*:link get_non_random_ability():const:client')
        func = parser.model.functions[0]
        self.assertEqual(func.return_type.type, 'DataAbility')
        self.assertEqual(func.return_type.is_link, True)


class TestParseClass(unittest.TestCase):

    def test_1(self):
        text = '''
        class Foo
        {
            int int_name
            float float_name
            Bar bar_name

            string:const:static INAPP = "inapp"

            function void function_1(int a){}
            function map<int, int> function_2(list<int> a){}
        }
        '''
        parser = Parser(Model())
        parser.parse_text(text)
        foo = parser.model.classes[0]
        self.assertEqual(foo.name, 'Foo')
        self.assertEqual(foo.type, 'class')
        self.assertEqual(len(foo.members), 4)

        self.assertEqual(foo.members[0].type, 'int')
        self.assertEqual(foo.members[0].name, 'int_name')

        self.assertEqual(foo.members[3].type, 'string')
        self.assertEqual(foo.members[3].name, 'INAPP')

    def test_2(self):
        text = '''
        class RewardsListHelper
        {
            list<DataReward*> rewards
        }
        '''
        parser = Parser(Model())
        parser.parse_text(text)
        foo = parser.model.classes[0]
        self.assertEqual(foo.name, 'RewardsListHelper')
        self.assertEqual(foo.type, 'class')
        self.assertEqual(len(foo.members), 1)

    def test_3(self):
        text = '''
        class RewardsListHelper
        {
            constructor()
            {
            }
        }
        '''
        parser = Parser(Model())
        parser.parse_text(text)
        foo = parser.model.classes[0]
        self.assertEqual(foo.name, 'RewardsListHelper')
        self.assertEqual(foo.type, 'class')
        self.assertIsNotNone(foo.constructor)
        self.assertEqual(foo.constructor.args, [])

    def test_4(self):
        text = '''
        class RewardsListHelper
        {
            constructor(int a=0, float b=1)
            {
            }
        }
        '''
        parser = Parser(Model())
        parser.parse_text(text)
        foo = parser.model.classes[0]
        self.assertEqual(foo.name, 'RewardsListHelper')
        self.assertEqual(foo.type, 'class')
        self.assertIsNotNone(foo.constructor)
        self.assertEqual(len(foo.constructor.args), 2)
        self.assertTrue(isinstance(foo.constructor.args[0][1], Object))
        self.assertTrue(isinstance(foo.constructor.args[1][1], Object))
        self.assertEqual(foo.constructor.args[0][1].type, 'int')
        self.assertEqual(foo.constructor.args[1][1].type, 'float')
        self.assertEqual(foo.constructor.args[0][0], 'a')
        self.assertEqual(foo.constructor.args[1][0], 'b')
        self.assertEqual(foo.constructor.args[0][1].initial_value, '0')
        self.assertEqual(foo.constructor.args[1][1].initial_value, '1')

    def test_5(self):
        text = '''
        class RewardsListHelper
        {
            int a
            float b
            
            constructor():generate
        }
        '''
        parser = Parser(Model())
        parser.parse_text(text)
        foo = parser.model.classes[0]
        self.assertEqual(foo.name, 'RewardsListHelper')
        self.assertEqual(foo.type, 'class')
        self.assertIsNotNone(foo.constructor)
        self.assertEqual(foo.constructor.name, 'constructor')
        self.assertEqual(foo.constructor.return_type.type, 'void')

        self.assertEqual(len(foo.constructor.args), 2)
        self.assertTrue(isinstance(foo.constructor.args[0][1], Object))
        self.assertTrue(isinstance(foo.constructor.args[1][1], Object))
        self.assertEqual(foo.constructor.args[0][1].type, 'int')
        self.assertEqual(foo.constructor.args[1][1].type, 'float')
        self.assertEqual(foo.constructor.args[0][0], 'a')
        self.assertEqual(foo.constructor.args[1][0], 'b')
        self.assertEqual(foo.constructor.args[0][1].initial_value, '0')
        self.assertEqual(foo.constructor.args[1][1].initial_value, '0.0')


class TestParseFunctionArgs(unittest.TestCase):

    def test_arg_default_params(self):
        parser = Parser(Model())
        text = 'function ReturnValue& test_9(int i=0)'
        parser.parse_text(text)
        func = parser.model.functions[0]
        self.assertTrue(isinstance(func, Function))
        self.assertEqual(func.name, 'test_9')
        self.assertEqual(len(func.args), 1)
        self.assertEqual(len(func.operations), 0)

        arg = func.args[0][1]
        self.assertTrue(isinstance(arg, Object))
        self.assertEqual(arg.type, 'int')
        self.assertEqual(arg.initial_value, '0')

    def test_arg_count_3(self):
        parser = Parser(Model())
        text = 'function ReturnValue& test_9(int a, int b, int c)'
        parser.parse_text(text)
        func = parser.model.functions[0]
        self.assertEqual(len(func.args), 3)

        self.assertTrue(isinstance(func.args[0][1], Object))
        self.assertEqual(func.args[0][0], 'a')
        self.assertEqual(func.args[0][1].type, 'int')

        self.assertTrue(isinstance(func.args[1][1], Object))
        self.assertEqual(func.args[1][0], 'b')
        self.assertEqual(func.args[1][1].type, 'int')

        self.assertTrue(isinstance(func.args[2][1], Object))
        self.assertEqual(func.args[2][0], 'c')
        self.assertEqual(func.args[2][1].type, 'int')


if __name__ == '__main__':
    unittest.main()



from ..core.class_ import Class


PREDEFINED = '''
#ifndef __@{namespace}_@{name}_h__
#define __@{namespace}_@{name}_h__
#include <assert.h>
#include <functional>
#include <set>
#include <thread>
#include <unordered_map>
#include "intrusive_ptr.h"

namespace @{namespace}
{
    class ObservableBase{
    public:
        virtual ~ObservableBase() = default;
        virtual void remove(void* object) = 0;
        virtual void remove(int object) = 0;
    };
    
    
    template<class R> class Observable;

    template<class R, class... A>
    class Observable<R(A...)> : public ObservableBase
    {
    public:
        Observable()
        : _threadId(std::this_thread::get_id())
        , _lock_counter(0)
        {
        }

        template<class T, class M, class... P>
        typename std::enable_if<std::is_member_function_pointer<M>::value>::type
        add(T object, M method, P&&... placeholders)
        {
            auto tag = get_tag(object);
            auto ptr = get_raw_pointer(object);
            if(is_locked())
                _listeners_to_add[tag] = std::bind(method, ptr, std::forward<P>(placeholders)...);
            else
                _listeners[tag] = std::bind(method, ptr, std::forward<P>(placeholders)...);
        }

        template<class T, class F>
        typename std::enable_if<!std::is_member_function_pointer<F>::value || std::is_function<F>::value>::type
        add(T object, F lambda)
        {
            auto tag = get_tag(object);
            if(is_locked())
                _listeners_to_add[tag] = lambda;
            else
                _listeners[tag] = lambda;
        }

        template<class ...T>
        void notify(T&&... args)
        {
            if(!is_locked())
            {
                lock();
                for(auto p : _listeners)
                {
                    if(_listeners_to_remove.count(p.first) == 0)
                        p.second(std::forward<T>(args)...);
                }
                unlock();
            }
        }

        virtual void remove(void* object) override
        {
            remove(get_tag(object));
        }
        virtual void remove(int object) override
        {
            remove(get_tag(object));
        }

    private:
        template<class T> long get_tag(intrusive_ptr<T> t)
        {
            return reinterpret_cast<long>(t.ptr());
        }
        template<class T> typename std::enable_if<std::is_pointer<T>::value, long>::type get_tag(T t)
        {
            return reinterpret_cast<long>(t);
        }
        template<class T> typename std::enable_if<std::is_integral<T>::value, long>::type get_tag(T t)
        {
            return static_cast<long>(t);
        }
        
        template<class T> T* get_raw_pointer(intrusive_ptr<T> t)
        {
            return t.ptr();
        }
        template<class T> typename std::enable_if<std::is_pointer<T>::value, T>::type get_raw_pointer(T t)
        {
            return t;
        }

        void remove(long tag)
        {
            auto iter = _listeners.find( tag );
            if( iter != _listeners.end() )
            {
                if(_lock_counter)
                    _listeners_to_remove.insert(tag);
                else
                    _listeners.erase( iter );
            }
        }
        bool is_locked() const
        {
            assert(_threadId == std::this_thread::get_id());
            return _lock_counter != 0;
        }

        void lock()
        {
            assert(_threadId == std::this_thread::get_id());
            ++_lock_counter;
        }

        void unlock()
        {
            assert(_threadId == std::this_thread::get_id());
            --_lock_counter;
            assert(_lock_counter >= 0);
            if(!is_locked())
            {
                _listeners.insert(_listeners_to_add.begin(), _listeners_to_add.end());
                for( auto func : _listeners_to_remove )
                {
                    remove(func);
                }
                _listeners_to_add.clear();
                _listeners_to_remove.clear();
            }
        }
    private:
        std::thread::id _threadId;
        int _lock_counter;
        std::unordered_map<long, std::function<R(A...)>> _listeners;
        std::unordered_map<long, std::function<R(A...)>> _listeners_to_add;
        std::set<long> _listeners_to_remove;
    };
}


#endif
'''


class GeneratorObserver(object):

    def __init__(self):
        pass

    @staticmethod
    def get_mock():
        cls = Class()
        cls.name = GeneratorObserver.get_observable_name()
        cls.type = 'class'
        cls.auto_generated = False
        return cls

    @staticmethod
    def get_observable_name():
        return 'Observable'

    @staticmethod
    def generate(model):
        text = PREDEFINED
        filename = GeneratorObserver.get_observable_name() + '.h'
        text = text.replace('@{namespace}', 'mg')
        text = text.replace('@{name}', GeneratorObserver.get_observable_name())

        mock = GeneratorObserver.get_mock()
        model.add_file(mock, filename, text)
        model.add_class(mock)



import re
from .model import SerializeFormat
from ..core.object import AccessSpecifier, Object


# pylint: disable=no-self-use
# pylint: disable=unused-argument
class WriterBase(object):

    def __init__(self, out_directory):
        self.model = None
        self.serializer = None
        self.out_directory = out_directory
        self.files = []
        self.created_files = []
        self.current_class = None

    def save(self, model):
        self.model = model
        for cls in model.classes:
            if not cls.auto_generated:
                continue
            self.current_class = cls
            sources = self.write_class(cls)
            for filename, content in sources:
                self.model.add_file(cls, filename, content)

    def write_class(self, cls):
        return [('', '')]

    def set_initial_values(self, cls):
        if cls.type == 'enum':
            for member in cls.members:
                if member.name == '_value' and member.initial_value is not None:
                    member.initial_value = cls.members[0].initial_value

    def get_constructor_data(self, cls):
        constructor_args = self.get_required_args_to_function(cls.constructor)
        constructor_body = ''
        if cls.constructor is not None:
            constructor_args = self.create_function_args_string(cls.constructor)
            constructor_body = cls.constructor.body
        if constructor_args is None:
            constructor_args = ''
        return constructor_args, constructor_body

    # Function methods
    def write_function(self, method):
        if method.name == 'constructor':
            return ''
        args = self.create_function_args_string(method)
        text = self.get_method_pattern(method).format(name=method.name,
                                                      args=args,
                                                      body=method.body,
                                                      access=AccessSpecifier.to_string(method.access),
                                                      class_name=self.current_class.name)
        if method.is_static:
            text = self.add_static_modifier_to_method(text)
        return text

    def create_function_args_string(self, method):
        args = []
        if self.get_required_args_to_function(method) is not None:
            args.append(self.get_required_args_to_function(method))
        for name, arg in method.args:
            pattern = self.get_method_arg_pattern(arg)
            typename = self.get_argument_typename(arg)
            args.append(pattern.format(name=name,
                                       type=typename,
                                       value=self.serializer.convert_initialize_value(arg.initial_value)))

        for template_type in method.template_types:
            obj = Object()
            obj.name = template_type
            obj.initial_value = self.get_nullptr_string()
            pattern = self.get_method_arg_pattern(obj)
            args.append(pattern.format(name=template_type, type='', value=self.get_nullptr_string()))

        args = ', '.join(args)
        return args

    # arg: Object
    def get_argument_typename(self, arg):
        typename = arg.type
        std_types = {
            'string': 'string',
            'int': 'int',
            'float': 'float',
            'bool': 'bool',
            'void': '',
        }
        if typename in std_types:
            return std_types[typename]
        if self.model.has_class(typename):
            return typename
        return ''

    def get_method_arg_pattern(self, obj):
        return '${name}={value}' if obj.initial_value is not None else '${name}'

    def get_method_pattern(self, method):
        return '{name}({args})\n{body}'

    def get_required_args_to_function(self, method):
        return None

    def get_nullptr_string(self):
        return 'null'

    def add_static_modifier_to_method(self, text):
        return 'static ' + text

    # End Function methods

    def write_object(self, obj):
        return ''

    def prepare_file(self, text):

        def remove_content(txt, obj_type):
            pattern = re.compile(r'\{\{format=%s\}\}[\s\S]+?\{\{end_format=%s\}\}' % (obj_type, obj_type))
            txt = pattern.sub('', txt)
            return txt

        def remove_marker(txt, obj_type):
            txt = txt.replace('{{format=%s}}' % obj_type, '')
            txt = txt.replace('{{end_format=%s}}' % obj_type, '')
            return txt

        all_formats = SerializeFormat.xml | SerializeFormat.json
        for format_code, format_string in SerializeFormat.get_all():
            if self.model.serialize_formats == all_formats:
                text = remove_content(text, format_string)
            else:
                if not (self.model.serialize_formats & format_code):
                    text = remove_content(text, format_string)
                else:
                    text = remove_marker(text, format_string)

        if self.model.serialize_formats != all_formats:
            text = remove_content(text, 'both')
        else:
            text = remove_marker(text, 'both')

        return text

    def prepare_file_codestype_php(self, text):
        text = WriterBase.prepare_file(self, text)
        tabs = 0
        lines = text.split('\n')
        text = list()

        def get_tabs(count):
            return '\t' * count

        for line in lines:
            line = line.strip()

            if line and line[0] == '}':
                tabs -= 1
            line = get_tabs(tabs) + line
            if line.strip() and line.strip()[0] == '{':
                tabs += 1
            text.append(line)
        text = '\n'.join(text)
        for i in range(10):
            tabs = '\n' + '\t' * i + '{'
            text = text.replace(tabs, ' {')
        return text



import sys
from .modifiers import Modifiers
from .object import Object, AccessSpecifier


class Function(object):

    def __init__(self):
        self.operations = []
        self.return_type = Object()
        self.name = ''
        self.args = []
        self.template_types = []
        self.is_const = False
        self.is_external = False
        self.is_static = False
        self.is_abstract = False
        self.is_template = False
        self.is_virtual = False
        self.is_friend = False
        self.is_generate = False
        self.side = 'both'
        self.access = AccessSpecifier.public
        self.body = ''
        self.translated = False
        self.specific_implementations = ''

    def get_return_type(self):
        return self.return_type

    def parse_method_body(self, body):
        operation_start_pos = 0
        curly_braces = 0
        square_brackets = 0

        for pos, char in enumerate(body):
            if char == '{':
                curly_braces += 1
            elif char == '}':
                curly_braces -= 1
            elif char in '([':
                square_brackets += 1
            elif char in ')]':
                square_brackets -= 1

            if curly_braces < 0 or square_brackets < 0:
                # TODO: Use Log.error
                print('error parsing function "{}" body'.format(self.name))
                sys.exit(-1)
            if not curly_braces and not square_brackets and char in ';}':
                operation = body[operation_start_pos:pos+1].strip()
                if operation:
                    self.operations.append(operation)
                operation_start_pos = pos + 1
                continue

        operation = body[operation_start_pos:].strip()
        if operation:
            self.operations.append(operation)

    def find_modifiers(self, string):
        if Modifiers.server in string:
            self.side = Modifiers.side_server
        if Modifiers.client in string:
            self.side = Modifiers.side_client
        self.is_external = self.is_external or Modifiers.external in string
        self.is_abstract = self.is_abstract or Modifiers.abstract in string
        self.is_static = self.is_static or Modifiers.static in string
        self.is_const = self.is_const or Modifiers.const in string
        self.is_virtual = self.is_virtual or Modifiers.virtual in string
        self.is_generate = self.is_virtual or Modifiers.generate in string

        string = AccessSpecifier.find_access(self, string)

        string = string.replace(Modifiers.server, '')
        string = string.replace(Modifiers.client, '')
        string = string.replace(Modifiers.external, '')
        string = string.replace(Modifiers.static, '')
        string = string.replace(Modifiers.const, '')
        string = string.replace(Modifiers.abstract, '')
        string = string.replace(Modifiers.virtual, '')
        string = string.replace(Modifiers.generate, '')
        return string





class Meta(object):
    __base__ = object

    def __init__(self, *args):
        self.args = args

    def build(self, value):
        if isinstance(self.args[0], Meta):
            return self.args[0].build(value)
        return self.args[0](value)



# -*- coding: utf-8 -*-
import xml.etree.ElementTree as ET
import json

from .AllTypesChildren import AllTypesChildren
from .common import *
from ..DeserializerXml import DeserializerXml
from ..SerializerJson import SerializerJson
from ..DeserializerJson import DeserializerJson
from ..Meta import Meta
from tests.test_serializer.py.gen.IntrusivePtr import IntrusivePtr, make_intrusive
from .TestEnum import TestEnum
from ..SerializerXml import SerializerXml


class AllTypes(object):
    TYPE = "AllTypes"
    __slots__ = ["int_value0", "int_value1", "int_value2", "float_value0", "float_value1", "bool_value0", "bool_value1", "str_value0", "str_value1", "int_list", "float_list", "bool_list", "string_list", "int_string_map", "float_string_map", "bool_string_map", "string_string_map", "string_int_map", "string_float_map", "string_bool_map", "object", "object_ptr", "object_list", "object_ptr_list", "object_map", "object_ptr_map", "enum_list", "enum_map"]

    def __init__(self, arg=None):
        self.int_value0 = 0 if arg is None else arg
        self.int_value1 = 0
        self.int_value2 = None
        self.float_value0 = 0
        self.float_value1 = 0.0
        self.bool_value0 = True
        self.bool_value1 = False
        self.str_value0 = ""
        self.str_value1 = ""
        self.int_list = []
        self.float_list = []
        self.bool_list = []
        self.string_list = []
        self.int_string_map = {}
        self.float_string_map = {}
        self.bool_string_map = {}
        self.string_string_map = {}
        self.string_int_map = {}
        self.string_float_map = {}
        self.string_bool_map = {}
        from .AllTypesChildren import AllTypesChildren
        self.object = AllTypesChildren()
        self.object_ptr = 0
        self.object_list = []
        self.object_ptr_list = []
        self.object_map = {}
        self.object_ptr_map = {}
        self.enum_list = []
        self.enum_map = {}

    def __hash__(self):
        return id(self)

    def initialize(self):
        from .AllTypesChildren import AllTypesChildren
        self.int_value0 = 1
        self.int_value1 = 1
        self.int_value2 = 0
        self.float_value0 = 1.0
        self.float_value1 = 1.0
        self.bool_value0 = False
        self.bool_value1 = True
        self.str_value0 = "test_string"
        self.str_value1 = "test_string"
        self.int_list.append(0)
        self.int_list.append(1)
        self.float_list.append(0.0)
        self.float_list.append(1.0)
        self.bool_list.append(True)
        self.bool_list.append(True)
        self.string_list.append("0")
        self.string_list.append("1")
        self.int_string_map[0] = "0"
        self.int_string_map[1] = "1"
        self.bool_string_map[True] = "0"
        self.bool_string_map[False] = "1"
        self.float_string_map[0.0] = "0"
        self.float_string_map[1.0] = "1"
        self.string_string_map["0"] = "0"
        self.string_string_map["1"] = "1"
        self.string_int_map["0"] = 0
        self.string_int_map["1"] = 1
        self.string_bool_map["0"] = True
        self.string_bool_map["1"] = False
        self.string_float_map["0"] = 0.0
        self.string_float_map["1"] = 1.0
        self.string_string_map["0"] = "0"
        self.string_string_map["1"] = "1"
        self.object.value = 0
        self.object_ptr = make_intrusive(AllTypesChildren)
        self.object_ptr.value = 0
        self.object_list.extend([self.object, self.object])
        self.enum_list.append(TestEnum.value1)
        self.enum_list.append(TestEnum.value2)
        self.enum_map[TestEnum.value1] = 1
        self.enum_map[TestEnum.value2] = 2
        pass

    @staticmethod
    def tests(logger):
        result = True
        inst = AllTypes()
        inst.initialize()
        return result
        pass

    def __eq__(self, rhs):
        result = True
        result = result and self.int_value0 == rhs.int_value0
        result = result and self.int_value1 == rhs.int_value1
        result = result and self.int_value2 == rhs.int_value2
        result = result and self.float_value0 == rhs.float_value0
        result = result and self.float_value1 == rhs.float_value1
        result = result and self.bool_value0 == rhs.bool_value0
        result = result and self.bool_value1 == rhs.bool_value1
        result = result and self.str_value0 == rhs.str_value0
        result = result and self.str_value1 == rhs.str_value1
        result = result and self.int_list == rhs.int_list
        result = result and self.float_list == rhs.float_list
        result = result and self.bool_list == rhs.bool_list
        result = result and self.string_list == rhs.string_list
        result = result and self.int_string_map == rhs.int_string_map
        result = result and self.float_string_map == rhs.float_string_map
        result = result and self.bool_string_map == rhs.bool_string_map
        result = result and self.string_string_map == rhs.string_string_map
        result = result and self.string_int_map == rhs.string_int_map
        result = result and self.string_float_map == rhs.string_float_map
        result = result and self.string_bool_map == rhs.string_bool_map
        result = result and self.object == rhs.object
        result = result and ((self.object_ptr is None and rhs.object_ptr is None) or (self.object_ptr is not None and rhs.object_ptr is not None and self.object_ptr == rhs.object_ptr))
        result = result and self.object_list == rhs.object_list
        result = result and self.object_ptr_list == rhs.object_ptr_list
        result = result and self.object_map == rhs.object_map
        result = result and self.object_ptr_map == rhs.object_ptr_map
        result = result and self.enum_list == rhs.enum_list
        result = result and self.enum_map == rhs.enum_map
        return result
        pass

    def __ne__(self, rhs):
        return not (self == rhs)
        pass

    def get_type(self):
        return AllTypes.TYPE
        pass

    def serialize_xml(self, serializer: SerializerXml):
        serializer.serialize(self.int_value0, 'int_value0')
        serializer.serialize(self.int_value1, 'int_value1')
        serializer.serialize(self.int_value2, 'int_value2')
        serializer.serialize(self.float_value0, 'float_value0')
        serializer.serialize(self.float_value1, 'float_value1')
        serializer.serialize(self.bool_value0, 'bool_value0', True)
        serializer.serialize(self.bool_value1, 'bool_value1', False)
        serializer.serialize(self.str_value0, 'str_value0')
        serializer.serialize(self.str_value1, 'str_value1')
        serializer.serialize(self.int_list, 'int_list')
        serializer.serialize(self.float_list, 'float_list')
        serializer.serialize(self.bool_list, 'bool_list')
        serializer.serialize(self.string_list, 'string_list')
        serializer.serialize(self.int_string_map, 'int_string_map')
        serializer.serialize(self.float_string_map, 'float_string_map')
        serializer.serialize(self.bool_string_map, 'bool_string_map')
        serializer.serialize(self.string_string_map, 'string_string_map')
        serializer.serialize(self.string_int_map, 'string_int_map')
        serializer.serialize(self.string_float_map, 'string_float_map')
        serializer.serialize(self.string_bool_map, 'string_bool_map')
        serializer.serialize(self.object, 'object')
        serializer.serialize(self.object_ptr, 'object_ptr')
        serializer.serialize(self.object_list, 'object_list')
        serializer.serialize(self.object_ptr_list, 'object_ptr_list')
        serializer.serialize(self.object_map, 'object_map')
        serializer.serialize(self.object_ptr_map, 'object_ptr_map')
        serializer.serialize(self.enum_list, 'enum_list')
        serializer.serialize(self.enum_map, 'enum_map')

    def deserialize_xml(self, deserializer: DeserializerXml):
        self.int_value0 = deserializer.deserialize('int_value0', int, 0)
        self.int_value1 = deserializer.deserialize('int_value1', int, 1)
        self.int_value2 = deserializer.deserialize('int_value2', int)
        self.float_value0 = deserializer.deserialize('float_value0', float, 0)
        self.float_value1 = deserializer.deserialize('float_value1', float, 0.0)
        self.bool_value0 = deserializer.deserialize('bool_value0', bool, True)
        self.bool_value1 = deserializer.deserialize('bool_value1', bool, False)
        self.str_value0 = deserializer.deserialize('str_value0', str, '')
        self.str_value1 = deserializer.deserialize('str_value1', str, '')
        self.int_list = deserializer.deserialize('int_list', Meta(list, int))
        self.float_list = deserializer.deserialize('float_list', Meta(list, float))
        self.bool_list = deserializer.deserialize('bool_list', Meta(list, bool))
        self.string_list = deserializer.deserialize('string_list', Meta(list, str))
        self.int_string_map = deserializer.deserialize('int_string_map', Meta(dict, int, str))
        self.float_string_map = deserializer.deserialize('float_string_map', Meta(dict, float, str))
        self.bool_string_map = deserializer.deserialize('bool_string_map', Meta(dict, bool, str))
        self.string_string_map = deserializer.deserialize('string_string_map', Meta(dict, str, str))
        self.string_int_map = deserializer.deserialize('string_int_map', Meta(dict, str, int))
        self.string_float_map = deserializer.deserialize('string_float_map', Meta(dict, str, float))
        self.string_bool_map = deserializer.deserialize('string_bool_map', Meta(dict, str, bool))
        self.object = deserializer.deserialize('object', AllTypesChildren)
        self.object_ptr = deserializer.deserialize('object_ptr', Meta(IntrusivePtr, AllTypesChildren))
        self.object_list = deserializer.deserialize('object_list', Meta(list, AllTypesChildren))
        self.object_ptr_list = deserializer.deserialize('object_ptr_list', Meta(list, Meta(IntrusivePtr, AllTypesChildren)))
        self.object_map = deserializer.deserialize('object_map', Meta(dict, str, AllTypesChildren))
        self.object_ptr_map = deserializer.deserialize('object_ptr_map', Meta(dict, str, Meta(IntrusivePtr, AllTypesChildren)))
        self.enum_list = deserializer.deserialize('enum_list', Meta(list, TestEnum))
        self.enum_map = deserializer.deserialize('enum_map', Meta(dict, TestEnum, int))


    def serialize_json(self, serializer: SerializerJson):
        serializer.serialize(self.int_value0, 'int_value0')
        serializer.serialize(self.int_value1, 'int_value1')
        # serializer.serialize(self.int_value2, 'int_value2')
        serializer.serialize(self.float_value0, 'float_value0')
        serializer.serialize(self.float_value1, 'float_value1')
        serializer.serialize(self.bool_value0, 'bool_value0')
        serializer.serialize(self.bool_value1, 'bool_value1')
        serializer.serialize(self.str_value0, 'str_value0')
        serializer.serialize(self.str_value1, 'str_value1')
        serializer.serialize(self.int_list, 'int_list')
        serializer.serialize(self.float_list, 'float_list')
        serializer.serialize(self.bool_list, 'bool_list')
        serializer.serialize(self.string_list, 'string_list')
        serializer.serialize(self.int_string_map, 'int_string_map')
        serializer.serialize(self.float_string_map, 'float_string_map')
        serializer.serialize(self.bool_string_map, 'bool_string_map')
        serializer.serialize(self.string_string_map, 'string_string_map')
        serializer.serialize(self.string_int_map, 'string_int_map')
        serializer.serialize(self.string_float_map, 'string_float_map')
        serializer.serialize(self.string_bool_map, 'string_bool_map')
        serializer.serialize(self.object, 'object')
        serializer.serialize(self.object_ptr, 'object_ptr')
        serializer.serialize(self.object_list, 'object_list')
        serializer.serialize(self.object_ptr_list, 'object_ptr_list')
        serializer.serialize(self.object_map, 'object_map')
        serializer.serialize(self.object_ptr_map, 'object_ptr_map')
        serializer.serialize(self.enum_list, 'enum_list')
        serializer.serialize(self.enum_map, 'enum_map')

    def deserialize_json(self, deserializer: DeserializerJson):
        self.int_value0 = deserializer.deserialize('int_value0', int, 0)
        self.int_value1 = deserializer.deserialize('int_value1', int, 1)
        self.int_value2 = deserializer.deserialize('int_value2', int)
        self.float_value0 = deserializer.deserialize('float_value0', float, 0)
        self.float_value1 = deserializer.deserialize('float_value1', float, 0.0)
        self.bool_value0 = deserializer.deserialize('bool_value0', bool, True)
        self.bool_value1 = deserializer.deserialize('bool_value1', bool, False)
        self.str_value0 = deserializer.deserialize('str_value0', str, '')
        self.str_value1 = deserializer.deserialize('str_value1', str, '')
        self.int_list = deserializer.deserialize('int_list', Meta(list, int))
        self.float_list = deserializer.deserialize('float_list', Meta(list, float))
        self.bool_list = deserializer.deserialize('bool_list', Meta(list, bool))
        self.string_list = deserializer.deserialize('string_list', Meta(list, str))
        self.int_string_map = deserializer.deserialize('int_string_map', Meta(dict, int, str))
        self.float_string_map = deserializer.deserialize('float_string_map', Meta(dict, float, str))
        self.bool_string_map = deserializer.deserialize('bool_string_map', Meta(dict, bool, str))
        self.string_string_map = deserializer.deserialize('string_string_map', Meta(dict, str, str))
        self.string_int_map = deserializer.deserialize('string_int_map', Meta(dict, str, int))
        self.string_float_map = deserializer.deserialize('string_float_map', Meta(dict, str, float))
        self.string_bool_map = deserializer.deserialize('string_bool_map', Meta(dict, str, bool))
        self.object = deserializer.deserialize('object', AllTypesChildren)
        self.object_ptr = deserializer.deserialize('object_ptr', Meta(IntrusivePtr, AllTypesChildren))
        self.object_list = deserializer.deserialize('object_list', Meta(list, AllTypesChildren))
        self.object_ptr_list = deserializer.deserialize('object_ptr_list', Meta(list, Meta(IntrusivePtr, AllTypesChildren)))
        self.object_map = deserializer.deserialize('object_map', Meta(dict, str, AllTypesChildren))
        self.object_ptr_map = deserializer.deserialize('object_ptr_map', Meta(dict, str, Meta(IntrusivePtr, AllTypesChildren)))
        self.enum_list = deserializer.deserialize('enum_list', Meta(list, TestEnum))
        self.enum_map = deserializer.deserialize('enum_map', Meta(dict, TestEnum, int))



import re
from ..base.save_plugin_base import SavePluginBase


class SavePlugin(SavePluginBase):
    def __init__(self, model):
        SavePluginBase.__init__(self, model)

    def _create_combine_file_header(self):
        header = '<?php\n'
        file_name = 'mg.php'
        return header, file_name

    def _finalize_combine_file(self, content):
        return content + '\n?>'

    def _remove_includes(self, file_content):
        file_content = file_content.replace('<?php\n', '')
        file_content = file_content.replace('\n?>', '')
        file_content = re.sub(r'\s*require_once.+', '', file_content)
        return file_content



# -*- coding: utf-8 -*-
from tests.test_serializer.py.gen.DataWrapper import DataWrapper
from tests.test_serializer.py.gen.IntrusivePtr import IntrusivePtr


class SerializerJson(object):
    def __init__(self, json):
        self.json = json

    def serialize(self, obj, key, default_value=None):
        if isinstance(obj, DataWrapper):
            self.serialize_attr(obj.name, key, default_value)
        elif isinstance(obj, IntrusivePtr):
            child = self.add_child(key)
            child.serialize_attr(obj.get_type(), 'type', default_value)
            obj.serialize_json(child)
            pass
        elif hasattr(obj, 'serialize_json'):
            obj.serialize_json(self.add_child(key))
        elif isinstance(obj, list):
            self.serialize_list(obj, key)
        elif isinstance(obj, dict):
            self.serialize_dict(obj, key)
        else:
            self.serialize_attr(obj, key, default_value)

    def serialize_attr(self, obj, key, default_value):
        if obj != default_value:
            self.json[key] = obj

    def serialize_dict(self, obj, key):
        js = SerializerJson(self.json) if not key else self.add_child_array(key)
        for k, v in obj.items():
            item = js.add_array_item('')
            item.serialize_key(k)
            item.serialize_value(v)

    def serialize_list(self, obj, key):
        js = SerializerJson(self.json) if not key else self.add_child_array(key)
        for item in obj:
            js.serialize_list_item(item)

    def add_child(self, key):
        if key:
            self.json[key] = {}
            return SerializerJson(self.json[key])
        else:
            return self

    def add_child_array(self, key):
        self.json[key] = []
        return SerializerJson(self.json[key])

    def add_array_item(self, key):
        self.json.append({})
        return SerializerJson(self.json[-1])

    def serialize_key(self, k):
        self.serialize(k, 'key')

    def serialize_value(self, v):
        self.serialize(v, 'value')

    def serialize_list_item(self, obj):
        self.json.append(None)
        if isinstance(obj, DataWrapper):
            self.json[-1] = obj.name
        elif isinstance(obj, IntrusivePtr):
            self.json[-1] = {}
            self.json[-1]['type'] = obj.get_type()
            obj.serialize_xml(SerializerJson(self.json[-1]))
            pass
        elif hasattr(obj, 'serialize_json'):
            self.json[-1] = {}
            serializer = SerializerJson(self.json[-1])
            serializer.serialize(obj, '')
        elif isinstance(obj, list):
            self.json[-1] = []
            SerializerJson(self.json[-1]).serialize_list(obj, '')
        elif isinstance(obj, dict):
            self.serialize_dict(obj, '')
        else:
            self.json[-1] = obj



from ..core.object import Object
from ..utils.error import Error
from .protocols import CPP_XML
from .protocols import CPP_JSON
from ..base.serializer_base import SerializerBase
from .writer import Writer

SERIALIZATION = 0
DESERIALIZATION = 1


class Serializer(SerializerBase):

    def __init__(self):
        SerializerBase.__init__(self)
        self.namespace = 'mg'

    def get_protocol(self, serialize_format):
        protocol = CPP_XML if serialize_format == 'xml' else CPP_JSON
        return protocol

    def get_parent_serialize_format(self):
        return '{}::{}({});\n'

    def get_serialization_function_args(self, serialize_type, serialize_format):
        def get_json():
            obj = Object()
            obj.type = 'SerializerJson' if serialize_type == SERIALIZATION else 'DeserializerJson'
            obj.is_ref = True
            return obj

        def get_xml():
            obj = Object()
            obj.type = 'SerializerXml' if serialize_type == SERIALIZATION else 'DeserializerXml'
            obj.is_ref = True
            return obj

        if serialize_format == 'json' and serialize_type == SERIALIZATION:
            return ['serializer', get_json()]
        if serialize_format == 'json' and serialize_type == DESERIALIZATION:
            return ['deserializer', get_json()]
        if serialize_format == 'xml' and serialize_type == SERIALIZATION:
            return ['serializer', get_xml()]
        if serialize_format == 'xml' and serialize_type == DESERIALIZATION:
            return ['deserializer', get_xml()]

    def build_serialize_operation(self, obj: Object, serialization_type, serialize_format):
        # return self.build_serialize_operation_(obj.name, obj.type, obj.initial_value,
        #                                        serialization_type, obj.template_args, obj.is_pointer,
        #                                        '', obj.is_link, serialize_format)
        result = ''
        if serialization_type == SERIALIZATION and obj.type in self.model.simple_types and obj.initial_value:
            result = 'serializer.serialize({name}, "{name}", {type}({value}));\n'
        elif serialization_type == SERIALIZATION:
            result = 'serializer.serialize({name}, "{name}");\n'
        elif serialization_type == DESERIALIZATION and obj.type in self.model.simple_types and obj.initial_value:
            result = 'deserializer.deserialize({name}, "{name}", {type}({value}));\n'
        elif serialization_type == DESERIALIZATION:
            result = 'deserializer.deserialize({name}, "{name}");\n'
        return result.format(name=obj.name, value=obj.initial_value, type=self.convert_type(obj.type))

    def build_serialize_operation_(self, obj_name, obj_type, obj_value, serialization_type, obj_template_args,
                                   obj_is_pointer, owner, is_link, serialize_format):
        index = 0
        if obj_value is None:
            index = 1
        type_ = obj_type
        if self.model.has_class(type_) and self.model.get_class(type_).type == 'enum':
            string = self._build_serialize_operation_enum(obj_name, serialization_type)
            return string

        if obj_type not in self.model.simple_types and type_ != 'list' and type_ != 'map':
            if is_link:
                type_ = 'link'
            elif obj_is_pointer:
                type_ = 'pointer'
            else:
                type_ = 'serialized'
        template_args = list()
        if obj_template_args:
            if type_ == 'map':
                if len(obj_template_args) != 2:
                    Error.exit(Error.MAP_TWO_ARGS, self.current_class.name, obj_name)
                if serialization_type == SERIALIZATION:
                    return self.build_map_serialization(obj_name, obj_template_args,
                                                        serialization_type, serialize_format)
                if serialization_type == DESERIALIZATION:
                    return self.build_map_deserialization(obj_name, obj_template_args, serialize_format)
            else:
                arg = obj_template_args[0]
                assert isinstance(arg, Object)
                assert isinstance(arg.type, str)
                arg_type = arg.type
                template_args.append(self.convert_type(arg_type))
                if arg.is_link:
                    type_ = 'list<link>'
                elif arg_type in self.model.simple_types:
                    type_ = '{0}<{1}>'.format(type_, arg_type)
                elif arg.is_pointer:
                    type_ = 'list<pointer>'
                elif self.model.has_class(arg.type) and self.model.get_class(arg.type).type == 'enum':
                    type_ = 'list<enum>'
                else:
                    type_ = '{0}<serialized>'.format(type_)
        if type_ not in self.serialize_protocol[serialization_type]:
            Error.exit(Error.UNKNOWN_SERIALISED_TYPE, type_, obj_type)
        pattern = self.serialize_protocol[serialization_type][type_][index]
        string = pattern.format(field=obj_name,
                                type=self.convert_type(obj_type),
                                default_value=obj_value,
                                arg_0=template_args[0] if template_args else '',
                                arg_1=template_args[1] if len(template_args) > 1 else '',
                                format=serialize_format,
                                namespace=self.namespace)
        return string + '\n\n'

    def build_map_serialization(self, obj_name, obj_template_args, serialization_type, serialize_format):
        key = obj_template_args[0]
        value = obj_template_args[1]
        assert isinstance(key, Object) and isinstance(value, Object)
        key_type = key.type
        value_type = value.type
        pattern = self.serialize_protocol[serialization_type]['map'][0]
        key_serialize = self.build_serialize_operation_('key', key_type, None, SERIALIZATION, [],
                                                        key.is_pointer, '', key.is_link, serialize_format)
        value_serialize = self.build_serialize_operation_('value', value_type, None, SERIALIZATION, value.template_args,
                                                          value.is_pointer, '', value.is_link, serialize_format)
        return pattern.format(field=obj_name,
                              key_serialize=key_serialize.strip(),
                              value_serialize=value_serialize.strip()) + '\n\n'

    def build_map_deserialization(self, obj_name, obj_template_args, serialize_format):
        key = obj_template_args[0]
        value = obj_template_args[1]
        assert isinstance(key, Object)
        assert isinstance(value, Object)
        assert isinstance(key.type, str)
        assert isinstance(value.type, str)
        key_type = key.type
        value_type = value.type
        pattern = self.serialize_protocol[DESERIALIZATION]['map'][0]
        if key.is_link:
            key_str = 'const {}* key(nullptr);'.format(key_type)
        elif key.is_pointer:
            key_str = 'auto key = make_intrusive<{}>();'.format(key_type)
        else:
            key_str = '{} key;'.format(self.convert_type(key_type))

        value_is_pointer = value.is_pointer
        key_serialize = self.build_serialize_operation_('key', key_type, None, DESERIALIZATION, [],
                                                        key.is_pointer, '', key.is_link, serialize_format)
        value_serialize = self.build_serialize_operation_('value', value_type, None, DESERIALIZATION,
                                                          value.template_args, value_is_pointer, '',
                                                          value.is_link, serialize_format)
        value_init = Writer.write_named_object(value, '', False, False)
        if value.is_pointer:
            value_init = 'intrusive_ptr<{}>'.format(value_type)
        return pattern.format(field=obj_name,
                              key_serialize=key_serialize.strip(),
                              value_serialize=value_serialize.strip(),
                              key=key_str,
                              value_type=value_init) + '\n\n'

    @staticmethod
    def convert_type(type_):
        types = {
            'list': 'std::vector',
            'map': 'std::map',
            'string': 'std::string',
        }
        if type_ in types:
            return types[type_]
        return type_

    def _build_serialize_operation_enum(self, obj_name, serialization_type):
        pattern = self.serialize_protocol[serialization_type]['enum'][0]
        return pattern.format(field=obj_name, namespace=self.namespace) + '\n\n'



from .constants import *
from ..base.writer_base import WriterBase
from ..base.model import SerializeFormat, Model
from ..core.class_ import Class


class GeneratorPredefinedFiles(object):

    def __init__(self):
        self.model = None

    def generate(self, model: Model):
        self.model = model
        writer = WriterBase('')
        writer.model = model

        self.build_common_content()

        for filename, content in FILES.items():
            model.add_file(None, filename, content)

    def build_common_content(self):
        content = COMMON
        xml = ''
        json = ''
        if self.model.serialize_formats & SerializeFormat.xml and self.model.serialize_formats & SerializeFormat.json:
            xml = COMMON_XML
            json = COMMON_JSON + CLONE_JSON
        elif self.model.serialize_formats & SerializeFormat.xml:
            xml = COMMON_XML + CLONE_XML
        elif self.model.serialize_formats & SerializeFormat.json:
            json = COMMON_JSON + CLONE_JSON
        content = content.replace('@{xml}', xml)
        content = content.replace('@{json}', json)
        FILES['common.php'] = content


FILES = {}


from .regex import RegexPatternJs
from ..base.translator_base import TranslatorBase


class Translator(TranslatorBase):

    def __init__(self):
        TranslatorBase.__init__(self)

    def replace_by_regex(self, func, body, model, args):
        args = ', '.join([x[0] for x in args])
        if not body and not args:
            return body
        body, strings = self.save_strings(body)
        for reg in RegexPatternJs.FUNCTION:
            body = self.replace(body, reg)
        for reg in RegexPatternJs.REPLACES:
            body = body.replace(reg[0], reg[1])
        body = self.restore_strings(body, strings)

        return body



import unittest

import os
import sys
import inspect

sys.path.insert(0, os.path.abspath(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))) + '/..'))
from mlc_tools.core.object import *
from mlc_tools.base.parser import Parser
from mlc_tools.base.model import Model


class TestParseModifiersTypes(unittest.TestCase):

    def test_1(self):
        obj = Parser.create_object('int:client')
        self.assertEqual(obj.side, 'client')

    def test_2(self):
        obj = Parser.create_object('int:server')
        self.assertEqual(obj.side, 'server')

    def test_3(self):
        obj = Parser.create_object('int*')
        self.assertTrue(obj.is_pointer)

    def test_4(self):
        obj = Parser.create_object('int&')
        self.assertTrue(obj.is_ref)

    def test_5(self):
        obj = Parser.create_object('int:runtime')
        self.assertTrue(obj.is_runtime)

    def test_6(self):
        obj = Parser.create_object('int:static')
        self.assertTrue(obj.is_static)

    def test_7(self):
        obj = Parser.create_object('int:const')
        self.assertTrue(obj.is_const)

    def test_8(self):
        obj = Parser.create_object('int:key')
        self.assertTrue(obj.is_key)

    def test_9(self):
        obj = Parser.create_object('int:link')
        self.assertTrue(obj.is_link)

    def test_10(self):
        obj = Parser.create_object('int:private')
        self.assertTrue(obj.access == AccessSpecifier.private)

    def test_11(self):
        obj = Parser.create_object('int:protected')
        self.assertTrue(obj.access == AccessSpecifier.protected)

    def test_12(self):
        obj = Parser.create_object('int:public')
        self.assertTrue(obj.access == AccessSpecifier.public)


class TestParseComplexTypes(unittest.TestCase):
    
    def test_1(self):
        obj = Parser.create_object('int key')
        self.assertEqual(obj.type, 'int')
        self.assertEqual(obj.name, 'key')
        self.assertEqual(obj.template_args, [])
        self.assertFalse(obj.is_pointer)
        self.assertFalse(obj.is_ref)
        self.assertFalse(obj.is_runtime)
        self.assertFalse(obj.is_static)
        self.assertFalse(obj.is_const)
        self.assertFalse(obj.is_key)
        self.assertFalse(obj.is_link)
        self.assertEqual(obj.side, 'both')
        self.assertEqual(obj.access, AccessSpecifier.public)

    def test_2(self):
        obj = Parser.create_object('DataStorage&:static')
        self.assertEqual(obj.type, 'DataStorage')
        self.assertEqual(obj.name, '')
        self.assertEqual(obj.template_args, [])
        self.assertFalse(obj.is_pointer)
        self.assertTrue(obj.is_ref)
        self.assertFalse(obj.is_runtime)
        self.assertTrue(obj.is_static)
        self.assertFalse(obj.is_const)
        self.assertFalse(obj.is_key)
        self.assertFalse(obj.is_link)
        self.assertEqual(obj.side, 'both')
        self.assertEqual(obj.access, AccessSpecifier.public)

    def test_3(self):
        obj = Parser.create_object('list<int>:static name')
        self.assertEqual(obj.type, 'list')
        self.assertEqual(obj.name, 'name')
        self.assertEqual(obj.template_args[0].type, 'int')
        self.assertFalse(obj.is_pointer)
        self.assertFalse(obj.is_ref)
        self.assertFalse(obj.is_runtime)
        self.assertTrue(obj.is_static)
        self.assertFalse(obj.is_const)
        self.assertFalse(obj.is_key)
        self.assertFalse(obj.is_link)
        self.assertEqual(obj.side, 'both')
        self.assertEqual(obj.access, AccessSpecifier.public)
    
    def test_4(self):
        obj = Parser.create_object('map<string, list<int>:static>:const name')
        self.assertEqual(obj.type, 'map')
        self.assertEqual(obj.name, 'name')
        self.assertEqual(obj.template_args[0].type, 'string')
        self.assertEqual(obj.template_args[1].type, 'list')
        self.assertTrue(obj.template_args[1].is_static)
        self.assertFalse(obj.is_pointer)
        self.assertFalse(obj.is_ref)
        self.assertFalse(obj.is_runtime)
        self.assertFalse(obj.is_static)
        self.assertTrue(obj.is_const)
        self.assertFalse(obj.is_key)
        self.assertFalse(obj.is_link)
        self.assertEqual(obj.side, 'both')
        self.assertEqual(obj.access, AccessSpecifier.public)


class TestParserOther(unittest.TestCase):

    def test_1(self):
        parser = Parser(Model())
        
        parser.model.language = 'cpp'
        self.assertTrue(parser.check_skip('function bool test(Logger* logger):static:py:php')[0])
        self.assertFalse(parser.check_skip('function bool test(Logger* logger):static:py:cpp')[0])
        self.assertFalse(parser.check_skip('function bool test(Logger* logger):static:cpp')[0])
        self.assertTrue(parser.check_skip('function bool test(Logger* logger):static:php')[0])
        self.assertTrue(parser.check_skip('function bool test(Logger* logger):static:py')[0])
        
        parser.model.language = 'py'
        self.assertFalse(parser.check_skip('function bool test(Logger* logger):static:py:php')[0])
        self.assertFalse(parser.check_skip('function bool test(Logger* logger):static:py:cpp')[0])
        self.assertTrue(parser.check_skip('function bool test(Logger* logger):static:cpp')[0])
        self.assertTrue(parser.check_skip('function bool test(Logger* logger):static:php')[0])
        self.assertFalse(parser.check_skip('function bool test(Logger* logger):static:py')[0])
        
        parser.model.language = 'php'
        self.assertFalse(parser.check_skip('function bool test(Logger* logger):static:py:php')[0])
        self.assertTrue(parser.check_skip('function bool test(Logger* logger):static:py:cpp')[0])
        self.assertTrue(parser.check_skip('function bool test(Logger* logger):static:cpp')[0])
        self.assertFalse(parser.check_skip('function bool test(Logger* logger):static:php')[0])
        self.assertTrue(parser.check_skip('function bool test(Logger* logger):static:py')[0])

    def test_2(self):
        parser = Parser(Model())
    
        parser.model.language = 'cpp'
        self.assertFalse(parser.check_skip('int:cpp test')[0])
        self.assertEqual(parser.check_skip('int:client:cpp test')[1], 'int:client test')
        
    def test_3(self):
        parser = Parser(Model())
    
        parser.model.language = 'cpp'
        parser.parse_text('list<int>:static:cpp name')
        obj = parser.model.objects[0]
        self.assertEqual(obj.type, 'list')
        self.assertEqual(obj.name, 'name')
    
        parser.model.language = 'py'
        parser.parse_text('list<int>:static:cpp name')
        obj = parser.model.objects[0]
        self.assertEqual(obj.type, 'list')
        self.assertEqual(obj.name, 'name')
        

if __name__ == '__main__':
    unittest.main()



import os
import sys
import glob
import inspect
import shutil
from time import sleep

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)

from mlc_tools import Mlc


def get_root():
    return os.path.abspath(os.path.dirname(os.path.abspath(__file__)) + '/..')


def run_tests(generator, root, withdata=False, cpp=True, python=True, php=True, js=True, with_join=False):
    def run(lang, serialized_format, join_to_one_file=False, clean_out_dir=True):
        out_directory = root + 'generated_%s' % (lang if lang != 'cpp' else lang + '/' + serialized_format)

        # clean python pyc files:
        def clean():
            try:
                shutil.rmtree(out_directory)
            except OSError:
                pass

        if clean_out_dir:
            clean()
        generator.generate(language=lang,
                           out_directory=out_directory,
                           formats=serialized_format,
                           join_to_one_file=join_to_one_file,
                           generate_ref_counter=True,
                           validate_allow_different_virtual_method=True
                           )
        if withdata:
            generator.generate_data(data_directory=root + 'data_%s/' % serialized_format,
                                    out_data_directory=root + 'assets')
        generator.run_test(test_script=root + 'test_%s.py' % lang,
                           test_script_args=serialized_format)
        print('-----------------------------------------')
        print('|  test with params [{}, {}] finished'.format(lang, serialized_format))
        print('-----------------------------------------')

    if cpp:
        generator.model.php_validate = False
        run('cpp', 'json', clean_out_dir=False)
        run('cpp', 'xml', clean_out_dir=False)
    if python:
        generator.model.php_validate = False
        run('py', 'json', False)
        run('py', 'xml', False)
        with_join and run('py', 'json', True)
        with_join and run('py', 'xml', True)
    if php:
        generator.model.php_validate = True
        run('php', 'json', False)
        run('php', 'xml', False)
        with_join and run('php', 'json', True)
        with_join and run('php', 'xml', True)
    if js:
        generator.model.php_validate = False
        run('js', 'json')


def simple_test():
    root = get_root() + '/tests/simple_test/'
    generator = Mlc(configs_directory=root + 'config',
                    side='client',
                    disable_logs=False,
                    generate_tests=True,
                    generate_intrusive=True,
                    generate_factory=True,
                    auto_registration=False,
                    user_includes=True)
    generator.add_config_directories(root + 'config_additional')
    generator.add_data_directories(root + 'data_additional')
    run_tests(generator, root, True, with_join=True)
    return 0


def test_functions():
    root = get_root() + '/tests/test_functions/'
    generator = Mlc(configs_directory=root + 'config', generate_intrusive=True, generate_factory=True)
    run_tests(generator, root)
    return 0


def test_database():
    root = get_root() + '/tests/test_database/'
    generator = Mlc(configs_directory=root + 'config',
                    generate_intrusive=True,
                    generate_factory=True,
                    generate_tests=True)
    run_tests(generator, root)
    return 0


def test_serialize():
    root = get_root()
    python = 'python3' if sys.version_info[0] == 3 else 'python'
    command = '{} {}/tests/test_serialize/run.py'.format(python, root)
    result = os.system(command)
    if 0 != result:
        sys.exit(1)
    return 0


def unit_tests_generator():
    root = get_root() + '/tests/unit_tests_generator/'
    generator = Mlc(configs_directory=root, generate_intrusive=True, generate_factory=True, generate_tests=True)
    run_tests(generator, root)
    return 0


def test_virtual_methods():
    root = get_root() + '/tests/test_virtual_methods/'
    generator = Mlc(configs_directory=root, generate_intrusive=True, generate_factory=True, generate_tests=True)
    run_tests(generator, root, python=False, php=False)
    return 0


def test_console():
    if sys.version_info[0] == 2:
        return
    python = 'python3'
    command = 'cd {1}/tests/test_console_compile; {0} run.py'.format(python, get_root())
    result = os.system(command)
    if 0 != result:
        sys.exit(1)
    return 0


if __name__ == '__main__':
    if len(sys.argv) == 1:
        simple_test()
        test_serialize()
        test_functions()
        unit_tests_generator()
        test_virtual_methods()
        test_console()
    else:
        exec(sys.argv[1])

    # Dont run this test in CI
    # test_database()
